import os
import time

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats
import statsmodels.sandbox.stats.multicomp
from fastcluster import linkage
from matplotlib import pyplot as plt
from scanpy import read  # So that we can call this function in cospar directly
from scipy.optimize import fmin
from scipy.spatial.distance import pdist
from sklearn.decomposition import PCA, TruncatedSVD
from sklearn.metrics import pairwise
from sklearn.neighbors import NearestNeighbors, kneighbors_graph
from tqdm import tqdm

from .. import logging as logg
from .. import plotting as pl
from .. import settings, tmap

# import scipy.stats


def get_dge_SW(ad, mask1, mask2, min_frac_expr=0.05, pseudocount=1):
    """
    Perform differential gene expression analysis.

    Parameters
    ----------
    ad: :class:`~anndata.AnnData` object
    mask1: `np.array`
        A np.array of `bool` for selecting group_1 cells.
    mask2: `np.array`
        A np.array of `bool` for selecting group_2 cells.
    min_frac_expr: `float`, optional (default: 0.05)
        Minimum expression fraction among selected states for a
        gene to be considered for DGE analysis.
    pseudocount: `int`, optional (default: 1)
        pseudo count for taking the gene expression ratio between the two groups

    Returns
    -------
    df: :class:`pandas.DataFrame`
        A pandas dataFrame, with columns: `gene`, `pv`, `mean_1`, `mean_2`, `ratio`
    """

    gene_mask = (
        (ad.X[mask1, :] > 0).sum(0).A.squeeze() / mask1.sum() > min_frac_expr
    ) | ((ad.X[mask2, :] > 0).sum(0).A.squeeze() / mask2.sum() > min_frac_expr)
    # print(gene_mask.sum())
    E1 = ad.X[mask1, :][:, gene_mask].toarray()
    E2 = ad.X[mask2, :][:, gene_mask].toarray()

    m1 = E1.mean(0) + pseudocount
    m2 = E2.mean(0) + pseudocount
    r = np.log2(m1 / m2)

    pv = np.zeros(gene_mask.sum())
    for ii, iG in enumerate(np.nonzero(gene_mask)[0]):
        pv[ii] = scipy.stats.ranksums(E1[:, ii], E2[:, ii])[1]
    pv = statsmodels.sandbox.stats.multicomp.multipletests(
        pv,
        alpha=0.05,
        method="fdr_bh",
    )[1]
    sort_idx = np.argsort(pv)

    df = pd.DataFrame(
        {
            "gene": ad.var_names.values.astype(str)[gene_mask][sort_idx],
            "Qvalue": pv[sort_idx],
            "mean_1": m1[sort_idx] - pseudocount,
            "mean_2": m2[sort_idx] - pseudocount,
            "ratio": r[sort_idx],
        }
    )

    return df


########## USEFUL SPARSE FUNCTIONS


def sparse_var(E, axis=0):
    """calculate variance across the specified axis of a sparse matrix"""

    mean_gene = E.mean(axis=axis).A.squeeze()
    tmp = E.copy()
    tmp.data **= 2
    return tmp.mean(axis=axis).A.squeeze() - mean_gene**2


def mean_center(E, column_means=None):
    """mean-center columns of a sparse matrix"""

    if column_means is None:
        column_means = E.mean(axis=0)
    return E - column_means


def normalize_variance(E, column_stdevs=None):
    """variance-normalize columns of a sparse matrix"""

    if column_stdevs is None:
        column_stdevs = np.sqrt(sparse_var(E, axis=0))
    return sparse_rowwise_multiply(E.T, 1 / column_stdevs).T


# this is not working well
def sparse_zscore(E, gene_mean=None, gene_stdev=None):
    """z-score normalize each column of a sparse matrix"""
    if gene_mean is None:
        gene_mean = E.mean(0)
    if gene_stdev is None:
        gene_stdev = np.sqrt(sparse_var(E))
    return sparse_rowwise_multiply((E - gene_mean).T, 1 / gene_stdev).T


def corr2_coeff(A, B):
    """
    Calculate Pearson correlation between matrix A and B

    A and B are allowed to have different shapes. This method
    does not work if A and B are constituted of constant row vectors,
    in which case the standard deviation becomes zero.

    Parameters
    ----------
    A: np.array, shape (cell_n0, gene_m)
    B: np.array, shape (cell_n1, gene_m)
    """

    resol = 10 ** (-15)
    # Rowwise mean of input arrays & subtract from input arrays themeselves
    A_mA = A - A.mean(1)[:, None]
    B_mB = B - B.mean(1)[:, None]

    # Sum of squares across rows
    ssA = (A_mA**2).sum(1)
    ssB = (B_mB**2).sum(1)

    # Finally get corr coeff
    return np.dot(A_mA, B_mB.T) / (np.sqrt(np.dot(ssA[:, None], ssB[None])) + resol)


def sparse_rowwise_multiply(E, a):
    """
    Multiply each row of the sparse matrix E by a scalar a

    Parameters
    ----------
    E: `np.array` or `sp.spmatrix`
    a: `np.array`
        A scalar vector.

    Returns
    -------
    Rescaled sparse matrix
    """

    nrow = E.shape[0]
    if nrow != a.shape[0]:
        logg.error("Dimension mismatch, multiplication failed")
        return E
    else:
        w = ssp.lil_matrix((nrow, nrow))
        w.setdiag(a)
        return w * E


def sparse_column_multiply(E, a):
    """
    Multiply each columns of the sparse matrix E by a scalar a

    Parameters
    ----------
    E: `np.array` or `sp.spmatrix`
    a: `np.array`
        A scalar vector.

    Returns
    -------
    Rescaled sparse matrix
    """

    ncol = E.shape[1]
    if ncol != a.shape[0]:
        logg.error("Dimension mismatch, multiplication failed")
        return E
    else:
        w = ssp.lil_matrix((ncol, ncol))
        w.setdiag(a)
        return ssp.csr_matrix(E) * w


# This is faster than v0
def matrix_row_or_column_thresholding(input_matrix, threshold=0.1, row_threshold=True):
    """
    Row or column-wise thresholding a matrix

    Set entries in a given row (column) to be zero, if its value is below threshold*max(row_vector).

    Parameters
    ----------
    input_matrix: `np.array`
    threshold: `float`, optional (default: 0.1)
    row_threshold: `bool`, optional (default: True)
        If true, perform row-wise thresholding; otherwise, column-wise.

    Returns
    -------
    Rescaled np.array matrix
    """

    # print("V1")
    # t1=time.time()
    if ssp.issparse(input_matrix):
        input_matrix = input_matrix.A
        # print("Turn the sparse matrix into numpy array")
        # print(f"Time-1: {time.time()-t1}")

    max_vector = np.max(input_matrix, int(row_threshold))
    for j in range(len(max_vector)):
        if row_threshold:
            idx = input_matrix[j, :] < threshold * max_vector[j]
            input_matrix[j, idx] = 0
        else:
            idx = input_matrix[:, j] < threshold * max_vector[j]
            input_matrix[idx, j] = 0
    # print(f"matrix_row_or_column_thresholding time:{time.time()-t1}")
    return input_matrix


# This is slower due to a step of copying
def matrix_row_or_column_thresholding_v0(
    input_matrix, threshold=0.1, row_threshold=True
):
    """
    Row or column-wise thresholding a matrix

    Set entries in a given row (column) to be zero, if its value is below threshold*max(row_vector).

    Parameters
    ----------
    input_matrix: `np.array`
    threshold: `float`, optional (default: 0.1)
    row_threshold: `bool`, optional (default: True)
        If true, perform row-wise thresholding; otherwise, column-wise.

    Returns
    -------
    Rescaled np.array matrix
    """
    print("V0")
    if ssp.issparse(input_matrix):
        input_matrix = input_matrix.A

    output_matrix = input_matrix.copy()
    max_vector = np.max(input_matrix, int(row_threshold))
    for j in range(len(max_vector)):
        # if j%2000==0: logg.hint(j)
        if row_threshold:
            idx = input_matrix[j, :] < threshold * max_vector[j]
            output_matrix[j, idx] = 0
        else:
            idx = input_matrix[:, j] < threshold * max_vector[j]
            output_matrix[idx, j] = 0

    return output_matrix


def get_pca(E, base_ix=[], numpc=50, keep_sparse=False, normalize=True, random_state=0):
    """
    Run PCA on the count matrix E, gene-level normalizing if desired.

    By default, it performs z-score transformation for each gene across all cells, i.e.,
    a gene normalization, before computing PCA. (There is currently no consensus on doing
    this or not. In scanpy, after count normalization (a per-cell normalization), it assumes
    that the individual gene counts in a cell is log-normally distributed, and performs a
    log-transformation before computing PCA. The z-score transformation is gene-specific,
    while the log-transformation is not.)

    Parameters
    ----------
    E: `sp.spmatrix`
        sparse count matrix
    base_ix: `np.array`
        List of column id's to sub-sample the matrix
    numpc: `int`, optional (default: 50)
        Number of principal components to keep
    keep_sparse: `bool`, optional (default: False)
        If true, do not subtract the mean, but just divide by
        standard deviation, before running PCA.
        If false, subtract the mean and then divide by standard deviation,
        thus performing Zscore transformation, before running PCA
    normalize: `bool`, optional (default: True)
        Perform Zscore transformation if keep_sparse=True,
        Otherwise, only rescale by the standard deviation.
    random_state: `int`, optional (default: 0)
        Random seed for PCA

    Returns
    -------
    PCA coordinates
    """

    # If keep_sparse is True, gene-level normalization maintains sparsity
    #     (no centering) and TruncatedSVD is used instead of normal PCA.

    if len(base_ix) == 0:
        base_ix = np.arange(E.shape[0])

    if keep_sparse:
        if normalize:  # normalize variance
            zstd = np.sqrt(sparse_var(E[base_ix, :]))
            Z = sparse_rowwise_multiply(E.T, 1 / zstd).T
        else:
            Z = E
        pca = TruncatedSVD(n_components=numpc, random_state=random_state)

    else:
        if normalize:
            zmean = E[base_ix, :].mean(0)
            zstd = np.sqrt(sparse_var(E[base_ix, :]))
            Z = sparse_rowwise_multiply((E - zmean).T, 1 / zstd).T
        else:
            Z = E
        pca = PCA(n_components=numpc, random_state=random_state)

    pca.fit(Z[base_ix, :])
    return pca.transform(Z)


########## GENE FILTERING


def runningquantile(x, y, p, nBins):
    """calculate the quantile of y in bins of x"""

    ind = np.argsort(x)
    x = x[ind]
    y = y[ind]

    dx = (x[-1] - x[0]) / nBins
    xOut = np.linspace(x[0] + dx / 2, x[-1] - dx / 2, nBins)

    yOut = np.zeros(xOut.shape)

    for i in range(len(xOut)):
        ind = np.nonzero((x >= xOut[i] - dx / 2) & (x < xOut[i] + dx / 2))[0]
        if len(ind) > 0:
            yOut[i] = np.percentile(y[ind], p)
        else:
            if i > 0:
                yOut[i] = yOut[i - 1]
            else:
                yOut[i] = np.nan

    return xOut, yOut


def get_vscores(E, min_mean=0, nBins=50, fit_percentile=0.1, error_wt=1):
    """
    Calculate v-score (above-Poisson noise statistic) for genes in the input sparse counts matrix
    Return v-scores and other stats
    """

    ncell = E.shape[0]

    mu_gene = E.mean(axis=0).A.squeeze()
    gene_ix = np.nonzero(mu_gene > min_mean)[0]
    mu_gene = mu_gene[gene_ix]

    tmp = E[:, gene_ix]
    tmp.data **= 2
    var_gene = tmp.mean(axis=0).A.squeeze() - mu_gene**2
    del tmp
    FF_gene = var_gene / mu_gene

    data_x = np.log(mu_gene)
    data_y = np.log(FF_gene / mu_gene)

    x, y = runningquantile(data_x, data_y, fit_percentile, nBins)
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]

    gLog = lambda input: np.log(input[1] * np.exp(-input[0]) + input[2])
    h, b = np.histogram(np.log(FF_gene[mu_gene > 0]), bins=200)
    b = b[:-1] + np.diff(b) / 2
    max_ix = np.argmax(h)
    c = np.max((np.exp(b[max_ix]), 1))
    errFun = lambda b2: np.sum(abs(gLog([x, c, b2]) - y) ** error_wt)
    b0 = 0.1
    b = fmin(func=errFun, x0=[b0], disp=False)
    a = c / (1 + b) - 1

    v_scores = FF_gene / ((1 + a) * (1 + b) + b * mu_gene)
    CV_eff = np.sqrt((1 + a) * (1 + b) - 1)
    CV_input = np.sqrt(b)

    return v_scores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b


def filter_genes(
    E,
    base_ix=[],
    min_vscore_pctl=85,
    min_counts=3,
    min_cells=3,
    show_vscore_plot=False,
    sample_name="",
):
    """
    Filter genes by expression level and variability

    Parameters
    ----------
    E: `sp.spmatrix`
        sparse count matrix
    base_ix: `np.array`
        List of column id's to sub-sample the matrix
    min_counts: int, optional (default: 3)
        Minimum number of UMIs per cell to be considered for selecting highly variable genes.
    min_cells: int, optional (default: 3)
        The minimum number of cells per gene to be considered for selecting highly variable genes.
    min_vscore_pctl: int, optional (default: 85)
        Genes wht a variability percentile higher than this threshold are marked as
        highly variable genes for dimension reduction. Range: [0,100]
    show_vscore_plot: `bool`, optional (default: False)
        If true, show the vscore plot for all genes
    sample_name: `str`, optional (default: '')
        Name of the plot title.

    Returns
    -------
    List of filtered gene indices (id's)
    """

    if len(base_ix) == 0:
        base_ix = np.arange(E.shape[0])

    Vscores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b = get_vscores(
        E[base_ix, :]
    )
    ix2 = Vscores > 0
    Vscores = Vscores[ix2]
    gene_ix = gene_ix[ix2]
    mu_gene = mu_gene[ix2]
    FF_gene = FF_gene[ix2]
    min_vscore = np.percentile(Vscores, min_vscore_pctl)
    ix = ((E[:, gene_ix] >= min_counts).sum(0).A.squeeze() >= min_cells) & (
        Vscores >= min_vscore
    )

    x_min = 0.5 * np.min(mu_gene)
    x_max = 2 * np.max(mu_gene)
    Mean_value = x_min * np.exp(np.log(x_max / x_min) * np.linspace(0, 1, 100))
    FanoFactor = (1 + a) * (1 + b) + b * Mean_value

    if show_vscore_plot:
        plt.figure(figsize=(4, 3))
        plt.scatter(
            np.log10(mu_gene), np.log10(FF_gene), c=[[0.8, 0.8, 0.8]], alpha=0.3, s=3
        )
        plt.scatter(
            np.log10(mu_gene)[ix], np.log10(FF_gene)[ix], c=[[0, 0, 0]], alpha=0.3, s=3
        )
        plt.plot(np.log10(Mean_value), np.log10(FanoFactor))
        plt.title(sample_name)
        plt.xlabel("log10(mean)")
        plt.ylabel("log10(Fano factor)")
        plt.show()

    if FanoFactor[-1] < FanoFactor[0]:
        logg.warn(
            f"The estimated Fano factor is NOT in expected form, which would affect the results.\n"
            "Please make sure that the count matrix adata.X is NOT log-transformed."
        )
    return gene_ix[ix]


# We found that this does not work
def remove_corr_genes(
    E, gene_list, exclude_corr_genes_list, test_gene_idx, min_corr=0.1
):
    """
    Remove signature-correlated genes from a list of test genes

    Parameters
    ----------
    E: ssp.csc_matrix, shape (n_cells, n_genes)
        full counts matrix
    gene_list: numpy array, shape (n_genes,)
        full gene list
    exclude_corr_genes_list: list of list(s)
        Each sublist is used to build a signature. Test genes correlated
        with this signature will be removed
    test_gene_idx: 1-D numpy array
        indices of genes to test for correlation with the
        gene signatures from exclude_corr_genes_list
    min_corr: float (default=0.1)
        Test genes with a Pearson correlation of min_corr or higher
        with any of the gene sets from exclude_corr_genes_list will
        be excluded

    Returns
    -------
    Numpy array of gene indices (a subset of test_gene_idx) that are not correlated with any of the gene signatures
    """

    seed_ix_list = []
    for l in exclude_corr_genes_list:
        seed_ix_list.append(
            np.array([i for i in range(len(gene_list)) if gene_list[i] in l], dtype=int)
        )

    exclude_ix = []
    for iSet in range(len(seed_ix_list)):
        seed_ix = seed_ix_list[iSet][
            E[:, seed_ix_list[iSet]].sum(axis=0).A.squeeze() > 0
        ]
        if type(seed_ix) is int:
            seed_ix = np.array([seed_ix], dtype=int)
        elif type(seed_ix[0]) is not int:
            seed_ix = seed_ix[0]
        indat = E[:, seed_ix]
        tmp = sparse_zscore(indat)
        tmp = tmp.sum(1).A.squeeze()

        c = np.zeros(len(test_gene_idx))
        for iG in range(len(c)):
            c[iG], _ = scipy.stats.pearsonr(tmp, E[:, test_gene_idx[iG]].A.squeeze())

        exclude_ix.extend(
            [test_gene_idx[i] for i in range(len(test_gene_idx)) if (c[i]) >= min_corr]
        )
    exclude_ix = np.array(exclude_ix)

    return np.array([g for g in test_gene_idx if g not in exclude_ix], dtype=int)


#################################################################

# check if a given id is in the list L2 (day 24 or 46), or L4 (day26)
# a conversion algorithm
def converting_id_from_fullSpace_to_subSpace(
    query_id_array_fullSpace, subSpace_id_array_inFull
):
    """
    Convert indices in the full space to those in the subspace.

    Parameters
    ----------
    query_id_array_fullSpace: `np.array` or `list`
        Indices in the full space
    subSpace_id_array_inFull: `np.array` or `list`
        Indices of a targeted sub population in the full space

    Returns
    -------
    np.array(query_id_inSub): `np.array`
        A converted np.array of indices in the subspace
    query_success: `np.array`
        A bool array of conversion success
    """

    id_sub = np.array(subSpace_id_array_inFull)
    query_id_inSub = []
    query_success = np.zeros(len(query_id_array_fullSpace), dtype=bool)
    # check one by one
    for j, id_full in enumerate(query_id_array_fullSpace):
        temp = np.nonzero(id_sub == id_full)[0]
        if len(temp) > 0:
            query_success[j] = True
            query_id_inSub.append(temp[0])

    return np.array(query_id_inSub), query_success


def converting_id_from_subSpace_to_fullSpace(
    query_id_array_subSpace, subSpace_id_array_inFull
):
    """
    Parameters
    ----------
    query_id_array_subSpace: `np.array` or `list`
        Indices in the sub space
    subSpace_id_array_inFull: `np.array` or `list`
        Indices of a targeted sub population in the full space

    Returns
    -------
    A converted np.array of indices in the full space
    """

    return np.array(subSpace_id_array_inFull)[query_id_array_subSpace]


def analyze_selected_fates(state_info, selected_fates):
    """
    Analyze selected fates.

    We return only valid fate clusters.

    Parameters
    ----------
    selected_fates: `list`
        List of selected fate clusters.
    state_info: `list`
        The state_info vector.

    Returns
    -------
    mega_cluster_list: `list`, shape (n_fate)
        The list of names for the mega cluster. This is relevant when
        `selected_fates` has a nested structure.
    valid_fate_list: `list`, shape (n_fate)
        It is the same as selected_fates, could contain a nested list
        of fate clusters. It screens for valid fates, though.
    fate_array_flat: `list` shape (>n_fate)
        List of all selected fate clusters. It flattens the selected_fates if it contains
        nested structure, and allows only valid clusters.
    sel_index_list: `list`, shape (n_fate)
        List of selected cell indexes for each merged cluster.
    """

    state_info = np.array(state_info)
    valid_state_annot = list(set(state_info))
    if type(selected_fates) is str:
        selected_fates = [selected_fates]
    if selected_fates is None:
        selected_fates = valid_state_annot

    fate_array_flat = []  # a flatten list of cluster names
    valid_fate_list = (
        []
    )  # a list of cluster lists, each cluster list is a macro cluster
    mega_cluster_list = []  # a list of string description for the macro cluster
    sel_index_list = []
    for xx in selected_fates:
        if type(xx) is list:
            valid_fate_list.append(xx)
            des_temp = ""
            temp_idx = np.zeros(len(state_info), dtype=bool)
            for zz in xx:
                if zz in valid_state_annot:
                    fate_array_flat.append(zz)
                    if des_temp != "":
                        des_temp = des_temp + "_"

                    des_temp = des_temp + str(zz)
                    temp_idx = temp_idx | (state_info == zz)
                else:
                    raise ValueError(
                        f"{zz} is not a valid cluster name. Please select from: {valid_state_annot}"
                    )
            mega_cluster_list.append(des_temp)
            sel_index_list.append(temp_idx)
        else:
            if xx in valid_state_annot:
                valid_fate_list.append([xx])

                fate_array_flat.append(xx)
                mega_cluster_list.append(str(xx))
            else:
                raise ValueError(
                    f"{xx} is not a valid cluster name. Please select from: {valid_state_annot}"
                )
                mega_cluster_list.append("")

            temp_idx = state_info == xx
            sel_index_list.append(temp_idx)

    # exclude invalid clusters
    mega_cluster_list = np.array(mega_cluster_list)
    fate_array_flat = np.array(fate_array_flat)
    sel_index_list = np.array(sel_index_list)
    valid_idx = mega_cluster_list != ""

    return (
        mega_cluster_list[valid_idx],
        valid_fate_list,
        fate_array_flat,
        sel_index_list[valid_idx],
    )


# v1, the new methods, more options.
def compute_shortest_path_distance(
    adata,
    num_neighbors_target=5,
    mode="distances",
    limit=np.inf,
    method="umap",
    normalize=True,
    use_existing_KNN_graph=False,
):
    """
    Compute shortest path distance from raw data.

    The distance matrix has two modes: 'connectivity' or 'distance'.
    We found that the 'connectivity' version is sensitive to local cell
    density heterogeneity, and the 'distance' version is more robust.
    This discrepancy might be due to that the KNN graph construction does not
    directly take into account local density heterogeneity.

    The default is the UMAP method, which takes into account local
    density heterogeneity into account when constructing the KNN graph.

    Parameters
    ----------
    adata: :class:`~anndata.AnnaData` object
    num_neighbors_target: `int`, optional (default: 5)
        Used to construct the KNN graph.
    mode: `str`, optional (default: 'distance')
        Options: {'distance','connectivity')
    limit: `float`, optional (default: np.inf)
        If the distance is about this, stop computation, and set
        the distance beyong this limist by `limit`. This can speed up computation.
    method: `str`, optional (default: 'umap')
        The method to construct the KNN graph. Options: {'umap','gauss','others'}.
        The frist two methods are based on sc.pp.neighbors, while the last is from
        kneighbors_graph.
    normalize: `bool`, optional (default: True)
        Normalize the distance matrix by its maximum value.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph.
        This overrides all other parameters.

    Returns
    -------
    The normalized distance matrix is returned.
    """

    from scanpy.neighbors import neighbors

    if (not use_existing_KNN_graph) or ("connectivities" not in adata.obsp.keys()):

        if mode != "connectivities":
            mode = "distances"

        logg.hint(f"Chosen mode is {mode}")
        if method == "umap":
            neighbors(adata, n_neighbors=num_neighbors_target, method="umap")
            adj_matrix = adata.obsp[mode]

        elif method == "gauss":
            neighbors(adata, n_neighbors=num_neighbors_target, method="gauss")
            adj_matrix = adata.obsp[mode]

        else:
            if mode == "distances":
                mode = "distance"
            else:
                mode = "connectivity"
            data_matrix = adata.obsm["X_pca"]
            adj_matrix = kneighbors_graph(
                data_matrix, num_neighbors_target, mode=mode, include_self=True
            )

    else:
        logg.info(
            f"Use existing KNN graph at adata.obsp[{mode}] for generating the smooth matrix"
        )
        adj_matrix = adata.obsp[mode]

    ShortPath_dis = ssp.csgraph.dijkstra(
        csgraph=ssp.csr_matrix(adj_matrix), directed=False, return_predecessors=False
    )
    ShortPath_dis_max = np.nanmax(ShortPath_dis[ShortPath_dis != np.inf])
    ShortPath_dis[
        ShortPath_dis > ShortPath_dis_max
    ] = ShortPath_dis_max  # set threshold for shortest paths

    # Set normalized cost matrices based on shortest paths matrices at target and source spaces
    if normalize:
        ShortPath_dis_final = ShortPath_dis / ShortPath_dis.max()
    else:
        ShortPath_dis_final = ShortPath_dis

    return ShortPath_dis_final


def add_neighboring_cells_to_a_map(initial_idx, adata, neighbor_N=5):
    """
    Add neighboring cells to an initially selected population

    Parameters
    ----------
    initial_idx: `np.array`, shape (n_cell,)
        A boolean array for state selection.
    adata: :class:`~anndata.AnnData` object, shape (n_cell, n_genes)
    neighbor_N: `int`, optional (default: 5)
        Number of neighbors for KNN graph construction.

    Returns
    -------
    post_idx: `np.array`, shape (n_cell,)
        A boolean array of selected cell states post expansion.
    """

    initial_idx = initial_idx > 0
    # print(f"Initial: {np.sum(initial_idx)}")
    #     if (np.sum(initial_idx)<size_thresh) & (np.sum(initial_idx)>0):
    #         #n0=np.round(size_thresh/np.sum(initial_idx))
    #         #sc.pp.neighbors(adata, n_neighbors=int(n0)) #,method='gauss')
    #         output_idx=adata.uns['neighbors']['connectivities'][initial_idx].sum(0).A.flatten()>0
    #         initial_idx=initial_idx | output_idx

    from scanpy.neighbors import neighbors

    neighbors(adata, n_neighbors=neighbor_N)  # ,method='gauss')
    output_idx = adata.obsp["connectivities"][initial_idx].sum(0).A.flatten() > 0
    post_idx = initial_idx | output_idx
    # print(f"Final: {np.sum(post_idx)}")

    return post_idx


def get_hierch_order(hm, dist_metric="euclidean", linkage_method="ward"):
    """
    This is used to order the barcode in generating the barcode heatmap.
    """
    np.random.seed(0)
    D = pdist(hm, dist_metric)
    Z = linkage(D, linkage_method)
    n = len(Z) + 1
    cache = dict()
    for k in range(len(Z)):
        c1, c2 = int(Z[k][0]), int(Z[k][1])
        c1 = [c1] if c1 < n else cache.pop(c1)
        c2 = [c2] if c2 < n else cache.pop(c2)
        cache[n + k] = c1 + c2
    o = np.array(cache[2 * len(Z)])
    return o


def above_the_line(x_array, x1, x2):
    """
    Return states above a specified line defined by (x1, x2).

    We assume that a state has only two coordinates.

    Parameters
    ----------
    x_array: `np.array`
        A 2-d matrix. Usually, an embedding for data points.
    x1: `np.array`
        A list or array of two entries.
    x2: `np.array`
        A list or array of two entries.

    Returns
    -------
    A boolean array.
    """

    return (x_array[:, 1] - x1[1]) > ((x2[1] - x1[1]) / (x2[0] - x1[0])) * (
        x_array[:, 0] - x1[0]
    )


def save_map(adata):
    """
    Save the adata and print the filename prefix.

    The filename prefix `data_des` will be printed, and
    the saved file can be accessed again using this prefix.
    """

    data_des = adata.uns["data_des"][-1]
    # data_path=adata.uns['data_path'][0]
    data_path = settings.data_path

    # need to remove these, otherwise, it does not work
    for xx in [
        "fate_trajectory",
        "multiTime_cell_id_t1",
        "multiTime_cell_id_t2",
        "fate_map",
        "fate_bias",
        "fate_potency",
    ]:
        if xx in adata.uns.keys():
            adata.uns.pop(xx)

    file_name = os.path.join(data_path, f"{data_des}_adata_with_transition_map.h5ad")
    adata.write_h5ad(file_name, compression="gzip")
    print(f"Saved file: {file_name}")


def smooth_a_vector(
    adata,
    vector,
    round_of_smooth=5,
    use_full_Smatrix=True,
    trunca_threshold=0.001,
    compute_new=False,
    neighbor_N=20,
    show_groups=True,
    save_subset=True,
):
    """ """

    vector = np.array(vector)
    if len(vector) != adata.shape[0]:
        logg.error("The length of the vector does not match that of adata.shape[0]")
        return None

    data_des = adata.uns["data_des"][0]
    data_des_1 = adata.uns["data_des"][-1]
    data_path = settings.data_path
    if "sp_idx" in adata.uns.keys():
        sp_idx = adata.uns["sp_idx"]
    else:
        sp_idx = np.ones(adata.shape[0]).astype(bool)

    # trunca_threshold=0.001 # this value is only for reducing the computed matrix size for saving
    temp_str = "0" + str(trunca_threshold)[2:]

    if use_full_Smatrix:
        similarity_file_name = os.path.join(
            data_path,
            f"{data_des}_Similarity_matrix_with_all_cell_states_kNN{neighbor_N}_Truncate{temp_str}",
        )

        if not os.path.exists(similarity_file_name + f"_SM{round_of_smooth}.npz"):
            logg.error(
                f"Similarity matrix at given parameters have not been computed before! Fiale name: {similarity_file_name}"
            )
            logg.error(
                f"Please use other Tmap inference function to build the full similarity matrix at the corresponding smooth rounds, using adata_orig."
            )
            return None

    else:
        similarity_file_name = os.path.join(
            data_path,
            f"{data_des_1}_Similarity_matrix_with_selected_states_kNN{neighbor_N}_Truncate{temp_str}",
        )

    # we cannot force it to compute new at this time. Otherwise, if we use_full_Smatrix, the resulting similarity is actually from adata, thus not full similarity.

    re_compute = (not use_full_Smatrix) and (
        compute_new
    )  # re-compute only when not using full similarity
    similarity_matrix_full = tmap.generate_similarity_matrix(
        adata,
        similarity_file_name,
        round_of_smooth=round_of_smooth,
        neighbor_N=neighbor_N,
        truncation_threshold=trunca_threshold,
        save_subset=save_subset,
        compute_new_Smatrix=re_compute,
    )

    if use_full_Smatrix:
        # pdb.set_trace()
        similarity_matrix_full_sp = similarity_matrix_full[sp_idx][:, sp_idx]

    else:
        similarity_matrix_full_sp = similarity_matrix_full

    smooth_vector = similarity_matrix_full_sp * vector
    if show_groups:
        x_emb = adata.obsm["X_emb"][:, 0]
        y_emb = adata.obsm["X_emb"][:, 1]
        fig_width = settings.fig_width
        fig_height = settings.fig_height
        fig = plt.figure(figsize=(2 * fig_width, fig_height))
        ax = plt.subplot(1, 2, 1)
        pl.customized_embedding(x_emb, y_emb, vector, ax=ax)
        ax = plt.subplot(1, 2, 2)
        pl.customized_embedding(x_emb, y_emb, smooth_vector, ax=ax)

    return smooth_vector


def update_time_ordering(adata, updated_ordering=None, mode="force"):
    """
    Update the ordering of time points at adata.uns['time_ordering']

    Parameters
    ----------
    updated_ordering: `list`, optional (default: none)
        A list of distinct time points in ascending order.
        If not provided, sort the time variable directly.
        However, these time variables are string. Their sorting
        may not be correct.
    mode: `str`
        Options: {'force','auto'}. In the 'auto' mode, the algorithm only
        update the ordering if 'time_ordering' has not been computed before.
        The default method 'force' will always update the ordering.
    """

    time_info = list(set(adata.obs["time_info"]))
    if updated_ordering is not None:
        N_match = np.sum(np.in1d(time_info, updated_ordering))
        if (len(updated_ordering) != N_match) or (
            len(updated_ordering) != len(time_info)
        ):
            logg.error(
                "The provided time points are not correct (wrong length, or invalid value)"
            )
            logg.info(
                f"Please provide an ordering of all time points in ascending order. Available time points are: {time_info}"
            )
            return None

    else:
        updated_ordering = np.sort(time_info)

    if mode == "auto":
        if "time_ordering" in adata.uns.keys():
            time_ordering_0 = adata.uns["time_ordering"]
            N_match = np.sum(np.in1d(time_info, time_ordering_0))
            if (len(time_ordering_0) != N_match) or (
                len(time_ordering_0) != len(time_info)
            ):
                logg.warn(
                    "Pre-computed time_ordering does not include the right time points. Re-compute it!"
                )
                adata.uns["time_ordering"] = np.array(updated_ordering)
                if updated_ordering is None:
                    logg.info(
                        f"Current time ordering from simple sorting: {updated_ordering}"
                    )
            else:
                # do not update
                return None
        else:
            adata.uns["time_ordering"] = np.array(updated_ordering)
            if updated_ordering is None:
                logg.info(
                    f"Current time ordering from simple sorting: {updated_ordering}"
                )
    else:
        # always update
        adata.uns["time_ordering"] = np.array(updated_ordering)
        if updated_ordering is None:
            logg.info(f"Current time ordering from simple sorting: {updated_ordering}")


def check_adata_structure(adata):
    """
    Check whether the adata has the right structure.
    """

    flag = True
    if not ("X_pca" in adata.obsm.keys()):
        logg.error("*X_pca* missing from adata.obsm")
        flag = False

    if not ("X_emb" in adata.obsm.keys()):
        logg.error("*X_emb* missing from adata.obsm")
        flag = False

    if not ("X_clone" in adata.obsm.keys()):
        logg.error("*X_clone* missing from adata.obsm")
        flag = False

    if not ("time_info" in adata.obs.keys()):
        logg.error("*time_info* missing from adata.obs")
        flag = False

    if not ("state_info" in adata.obs.keys()):
        logg.error("*state_info* missing from adata.obs")
        flag = False

    if flag:
        print("The adata structure looks fine!")


def save_preprocessed_adata(adata, data_des=None):
    """
    Save preprocessed adata.

    It will remove unneeded entries, and use the default
    prefix (data_des) to save the results if a new data_des is not provided.
    """

    if data_des is None:
        data_des = adata.uns["data_des"][0]

    data_path = settings.data_path

    for xx in [
        "fate_trajectory",
        "multiTime_cell_id_t1",
        "multiTime_cell_id_t2",
        "fate_map",
        "fate_bias",
        "fate_potency",
        "transition_map",
        "intraclone_transition_map",
        "clonal_transition_map",
        "OT_transition_map",
        "HighVar_transition_map",
        "Tmap_cell_id_t1",
        "Tmap_cell_id_t2",
        "available_map",
        "clonal_cell_id_t1",
        "clonal_cell_id_t2",
    ]:
        if xx in adata.uns.keys():
            adata.uns.pop(xx)
        if xx in adata.obs.keys():
            adata.obs.pop(xx)
        if xx in adata.obsm.keys():
            adata.obsm.pop(xx)

    # check_list=list(adata.obsm.keys())
    # for xx in  check_list:
    #     if xx not in ['X_clone', 'X_emb', 'X_pca']:
    #         adata.obsm.pop(xx)

    # check_list=list(adata.obs.keys())
    # for xx in  check_list:
    #     if xx not in ['state_info', 'time_info']:
    #         adata.obs.pop(xx)

    file_name = os.path.join(data_path, f"{data_des}_adata_preprocessed.h5ad")
    adata.write_h5ad(file_name, compression="gzip")
    print(f"Saved file: {file_name}")


def load_saved_adata_with_key(data_des):
    """
    Load pre-saved adata based on the prefix 'data_des'
    """

    data_path = settings.data_path
    # print(f"Load data: data_des='{data_des}'")
    file_name = os.path.join(data_path, f"{data_des}_adata_with_transition_map.h5ad")
    if os.path.exists(file_name):
        adata = read(file_name)
        return adata
    else:
        logg.error(f"The file does not existed yet")


def check_available_map(adata):
    """
    Check available transition map.

    Update adata.uns['available_map'].
    """

    available_map = []
    for xx in adata.uns.keys():
        if "transition_map" in xx:
            available_map.append(xx)
    adata.uns["available_map"] = available_map


def switch_adata_representation(adata, to_new=True):
    if to_new:
        adata.obsm["X_clone"] = adata.obsm["cell_by_clone_matrix"]
        adata.obs["state_info"] = adata.obs["state_annotation"]
        # adata.uns['data_des']=['paper_OneTimeClone_t*pos_17*pos_21*D27']
        adata.obsm.pop("cell_by_clone_matrix")
        adata.obs.pop("state_annotation")
    else:
        adata.obsm["cell_by_clone_matrix"] = adata.obsm["X_clone"]
        adata.obs["state_annotation"] = adata.obs["state_info"]
        # adata.uns['data_des']=['paper_OneTimeClone_t*pos_17*pos_21*D27']
        adata.obsm.pop("X_clone")
        adata.obs.pop("state_info")


def selecting_cells_by_time_points(time_info, selected_time_points):
    """
    Check validity of selected time points, and return the selected index.

    selected_time_points can be either a string or a list.

    If selected_time_points=[], we select all cell states.
    """

    time_info = np.array(time_info)
    valid_time_points = set(time_info)
    if selected_time_points is not None:
        if type(selected_time_points) is str:
            selected_times = [selected_time_points]
        else:
            selected_times = list(selected_time_points)

        sp_idx = np.zeros(len(time_info), dtype=bool)
        for xx in selected_times:
            if xx not in valid_time_points:
                logg.error(f"{xx} is not a valid time point.")
            sp_id_temp = np.nonzero(time_info == xx)[0]
            sp_idx[sp_id_temp] = True
    else:
        sp_idx = np.ones(len(time_info), dtype=bool)

    return sp_idx


def check_available_clonal_info(adata):

    X_clone = adata.obsm["X_clone"]
    time_info = adata.obs["time_info"]

    update_time_ordering(adata, mode="auto")

    # record time points with clonal information
    if ssp.issparse(X_clone):
        clone_N_per_cell = X_clone.sum(1).A.flatten()
    else:
        clone_N_per_cell = X_clone.sum(1)

    clonal_time_points = []
    for xx in list(set(time_info)):
        idx = np.array(time_info) == xx
        if np.sum(clone_N_per_cell[idx]) > 0:
            clonal_time_points.append(xx)

    time_ordering = adata.uns["time_ordering"]
    sel_idx_temp = np.in1d(time_ordering, clonal_time_points)
    clonal_time_points = time_ordering[sel_idx_temp]
    adata.uns["clonal_time_points"] = clonal_time_points


def check_available_choices(adata):
    """
    Check available parameter choices.

    Also update adata.uns['available_map'] and adata.uns['clonal_time_points'].
    """

    check_available_map(adata)
    available_map = adata.uns["available_map"]

    check_available_clonal_info(adata)
    clonal_time_points = adata.uns["clonal_time_points"]

    print("Available transition maps:", available_map)

    if "state_info" in adata.obs.keys():
        print("Available clusters:", list(set(adata.obs["state_info"])))
    else:
        print("No state_info yet")

    if "time_ordering" in adata.uns.keys():
        print("Available time points:", adata.uns["time_ordering"])
    else:
        print("No time_ordering yet")
    print("Clonal time points:", clonal_time_points)


def compute_pca(m1, m2, n_components):
    matrices = list()
    matrices.append(m1 if not ssp.isspmatrix(m1) else m1.toarray())
    matrices.append(m2 if not ssp.isspmatrix(m2) else m2.toarray())
    x = np.vstack(matrices)
    mean_shift = x.mean(axis=0)
    x = x - mean_shift
    n_components = min(n_components, x.shape[0])  # n_components must be <= ncells
    pca = PCA(n_components=n_components, random_state=58951)
    pca.fit(x.T)
    comp = pca.components_.T
    m1_len = m1.shape[0]
    m2_len = m2.shape[0]
    pca_1 = comp[0:m1_len]
    pca_2 = comp[m1_len : (m1_len + m2_len)]
    return pca_1, pca_2, pca, mean_shift


def compute_default_cost_matrix(a, b, eigenvals=None):

    if eigenvals is not None:
        a = a.dot(eigenvals)
        b = b.dot(eigenvals)

    cost_matrix = pairwise.pairwise_distances(
        a.toarray() if ssp.isspmatrix(a) else a,
        b.toarray() if ssp.isspmatrix(b) else b,
        metric="sqeuclidean",
        n_jobs=-1,
    )
    cost_matrix = cost_matrix / np.median(cost_matrix)
    return cost_matrix


def compute_gene_exp_distance(adata, p0_indices, p1_indices, pc_n=30):
    """
    Compute the gene expression distance between t0 and t1 states.

    p0_indices could be either a boolean array or index array
    """
    p0 = adata[p0_indices, :]
    p1 = adata[p1_indices, :]
    p0_x, p1_x, pca, mean = compute_pca(p0.X, p1.X, pc_n)
    eigenvals = np.diag(pca.singular_values_)
    # gene_exp_dis_t0 = compute_default_cost_matrix(p0_x, p0_x, eigenvals)
    # gene_exp_dis_t1 = compute_default_cost_matrix(p1_x, p1_x, eigenvals)
    gene_exp_dis_t0t1 = compute_default_cost_matrix(p0_x, p1_x, eigenvals)
    return gene_exp_dis_t0t1


def update_data_description(adata, append_info=None, data_des=None):
    """
    Update data_des, a string to distinguish different datasets
    """

    if "data_des" not in adata.uns.keys():
        if data_des is None:
            logg.warn("data_des not set. Set to be [cospar]")
            adata.uns["data_des"] = ["cospar"]
        else:
            adata.uns["data_des"] = [data_des]
    else:
        if data_des is not None:
            adata.uns["data_des"][-1] = data_des

    if append_info is not None:
        data_des_0 = adata.uns["data_des"][-1]
        data_des_1 = f"{data_des_0}_{append_info}"
        adata.uns["data_des"][-1] = data_des_1


def set_up_folders(data_path_new=None, figure_path_new=None):
    from pathlib import Path, PurePath

    if data_path_new is not None:
        settings.data_path = data_path_new
    if figure_path_new is not None:
        settings.figure_path = figure_path_new

    if not Path(settings.data_path).is_dir():
        logg.info(f"creating directory {settings.data_path} for saving data")
        Path(settings.data_path).mkdir(parents=True)

    if not Path(settings.figure_path).is_dir():
        logg.info(f"creating directory {settings.figure_path} for saving figures")
        Path(settings.figure_path).mkdir(parents=True)


def get_X_clone_with_reference_ordering(
    clone_data_cell_id,
    clone_data_barcode_id,
    reference_cell_id=None,
    reference_clone_id=None,
    drop_duplicates=True,
    count_value=None,
):
    """
    Build the X_clone matrix from data.

    Convert the raw clonal data table (long format): [clone_data_cell_id,clone_data_barcode_id]
    to X_clone (wide format) based on the reference_cell_id.

    Parameters
    ----------
    clone_data_cell_id: `list`
        The list of cell id for each corresponding sequenced barcode.
    clone_data_barcode_id: `list`
        The list of barcode id from sequencing. It has the same shape as clone_data_cell_id.
    reference_cell_id: `list`
        A list of uniuqe cell id. X_clone will be generated based on its cell id ordering.
        This has to be provided to match the cell ordering in the adata object.
    reference_clone_id: `list`, optional (default: None)
        A list of uniuqe clone id. If provided, X_clone will be generated based on its barcode ordering.
    count_value: `list`, optional (default: None)
        A list of count values corresponding to clone_data_cell_id. If not set, it is default to 1.

    Returns
    -------
    X_clone: `ssp.sparse`
        The clonal data matrix, with the row in cell id, and column in barcode id.
    reference_clone_id: `list`
    """

    if reference_clone_id is None:
        reference_clone_id = list(set(clone_data_barcode_id))
    if reference_cell_id is None:
        reference_cell_id = list(set(clone_data_cell_id))
    if count_value is None:
        count_value = np.ones(len(clone_data_cell_id))

    reference_clone_id = np.array(reference_clone_id)
    clone_map = {x: j for j, x in enumerate(reference_clone_id)}
    reference_cell_id = np.array(reference_cell_id)
    cell_map = {x: j for j, x in enumerate(reference_cell_id)}
    count_value = np.array(count_value)

    df = pd.DataFrame(
        {"cell_id": clone_data_cell_id, "clone_id": clone_data_barcode_id}
    )
    if reference_cell_id is not None:
        df = df[(df["cell_id"].isin(reference_cell_id))]

    if reference_clone_id is not None:
        df = df[(df["clone_id"].isin(reference_clone_id))]

    if drop_duplicates:
        df = df.drop_duplicates()

    clone_data_cell_id = df["cell_id"]
    clone_data_barcode_id = df["clone_id"]
    clone_data_cell_id = np.array(clone_data_cell_id)
    clone_data_barcode_id = np.array(clone_data_barcode_id)

    X_clone_row = []
    X_clone_col = []
    X_clone_val = []
    for j in tqdm(range(len(clone_data_cell_id))):
        X_clone_row.append(cell_map[clone_data_cell_id[j]])
        X_clone_col.append(clone_map[clone_data_barcode_id[j]])
        X_clone_val.append(count_value[j])

    X_clone = ssp.coo_matrix(
        (X_clone_val, (X_clone_row, X_clone_col)),
        shape=(len(reference_cell_id), len(reference_clone_id)),
    )
    X_clone = ssp.csr_matrix(X_clone)

    sp_idx = X_clone.sum(0).A.flatten() > 0
    return X_clone[:, sp_idx], reference_clone_id[sp_idx], reference_cell_id


def parse_output_choices(adata, key_word, where="obs", interrupt=True):
    if where == "obs":
        raw_choices = [xx for xx in adata.obs.keys() if xx.startswith(f"{key_word}")]
    else:
        raw_choices = [xx for xx in adata.uns.keys() if xx.startswith(f"{key_word}")]

    if (interrupt) and (len(raw_choices) == 0):
        raise ValueError(
            f"{key_word} has not been computed yet. Please run the counterpart function at cs.tl.XXX using the appropriate source name."
        )

    available_choices = []
    for xx in raw_choices:
        y = xx.split(f"{key_word}")[1]
        if y.startswith("_"):
            y = y[1:]
        available_choices.append(y)

    return available_choices


def rename_list(old_names, new_names):
    if new_names is None:
        new_names = old_names

    if len(old_names) != len(new_names):
        logg.warn(
            "The new name list does not have the same length as old name list. Rename operation aborted"
        )
        new_names = old_names
    return new_names


def check_input_parameters(adata, **kwargs):
    keys = kwargs.keys()
    check_available_clonal_info(adata)
    clonal_time_points_0 = np.array(adata.uns["clonal_time_points"])
    time_ordering = np.array(adata.uns["time_ordering"])
    if len(time_ordering) == 1:
        raise ValueError(
            "There is only one time point. Tmap inference requires at least 2 time points. Inference aborted."
        )

    if ("save_subset" in keys) and ("smooth_array" in keys):
        save_subset = kwargs["save_subset"]
        smooth_array = kwargs["smooth_array"]
        if save_subset:
            if not (
                np.all(np.diff(smooth_array) <= 0)
                and np.all(np.array(smooth_array) % 5 == 0)
            ):
                raise ValueError(
                    "The smooth_array contains numbers not multiples of 5 or not in descending order.\n"
                    "The correct form is like [20,15,10], or [10,10,10,5]."
                    "You can also set save_subset=False to explore arbitrary smooth_array structure."
                )

    if "data_des" not in adata.uns.keys():
        logg.warn("data_des not defined at adata.uns['data_des']. Set to be 'cospar'")
        adata.uns["data_des"] = ["cospar"]

    if "later_time_point" in keys:
        later_time_point = kwargs["later_time_point"]
        if (later_time_point is not None) and (
            later_time_point not in clonal_time_points_0
        ):
            raise ValueError(
                f"later_time_point is not all among {clonal_time_points_0}. Computation aborted!"
            )

    if "initial_time_points" in keys:
        initial_time_points = kwargs["initial_time_points"]
        N_valid_time = np.sum(np.in1d(time_ordering, initial_time_points))
        if (N_valid_time != len(initial_time_points)) or (N_valid_time < 1):
            raise ValueError(
                f"The 'initial_time_points' are not all valid. Please select from {time_ordering}"
            )

    if "clonal_time_points" in keys:
        clonal_time_points = kwargs["clonal_time_points"]
        N_valid_time = np.sum(np.in1d(clonal_time_points_0, clonal_time_points))
        if (N_valid_time != len(clonal_time_points)) or (N_valid_time < 2):
            raise ValueError(
                f"Selected time points are not all among {clonal_time_points_0}, or less than 2 time points are selected. Computation aborted!"
            )



"""KAILINss - dynamic inference by integrating transcriptome and lineage information"""

import cospar as cs
import numpy as np
import os

from . import workcatalogue
from . import preprocessing
from . import kl_sample
from . import omic_functions
from . import database
from .kl_settings import *
from . import test_others
from . import database
__version__ = "1.0.0"




import os
import time

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from numpy.lib.twodim_base import tril_indices
from scipy.cluster import hierarchy

# from plotnine import *
from sklearn.manifold import SpectralEmbedding

from cospar import tool as tl
from cospar.plotting import _utils as pl_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings


def gene_expression_dynamics(
    adata,
    selected_fate,
    gene_name_list,
    traj_threshold=0.1,
    source="transition_map",
    invert_PseudoTime=False,
    mask=None,
    compute_new=True,
    gene_exp_percentile=99,
    n_neighbors=8,
    plot_raw_data=False,
    stat_smooth_method="loess",
    ggplot_font_size=11,
):
    """
    Plot gene trend along the inferred dynamic trajectory.

    The results should be pre-computed from :func:`cospar.tl.progenitor` or
    :func:`cospar.tl.iterative_differentiation`

    Using the states that belong to the trajectory, it computes the pseudo time
    for these states and shows expression dynamics of selected genes along
    this pseudo time.

    Specifically, we first construct KNN graph, compute spectral embedding,
    and take the first component as the pseudo time. To create dynamics for a
    selected gene, we re-weight the expression of this gene at each cell by its
    probability belonging to the trajectory, and rescale the expression at selected
    percentile value. Finally, we fit a curve to the data points.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_fate: `str`, or `list`
        targeted cluster of the trajectory, as consistent with adata.obs['state_info']
        When it is a list, the listed clusters are combined into a single fate cluster.
    gene_name_list: `list`
        List of genes to plot on the dynamic trajectory.
    traj_threshold: `float`, optional (default: 0.1), range: (0,1)
        Relative threshold, used to thresholding the inferred dynamic trajecotry to select states.
    invert_PseudoTime: `bool`, optional (default: False)
        If true, invert the pseudotime: 1-pseuotime. This is useful when the direction
        of pseudo time does not agree with intuition.
    mask: `np.array`, optional (default: None)
        A boolean array for further selecting cell states.
    compute_new: `bool`, optional (default: True)
        If true, compute everyting from stratch (as we save computed pseudotime)
    gene_exp_percentile: `int`, optional (default: 99)
        Plot gene expression below this percentile.
    n_neighbors: `int`, optional (default: 8)
        Number of nearest neighbors for constructing KNN graph.
    plot_raw_data: `bool`, optional (default: False)
        Plot the raw gene expression values of each cell along the pseudotime.
    stat_smooth_method: `str`, optional (default: 'loess')
        Smooth method used in the ggplot. Current available choices are:
        'auto' (Use loess if (n<1000), glm otherwise),
        'lm' (Linear Model),
        'wls' (Linear Model),
        'rlm' (Robust Linear Model),
        'glm' (Generalized linear Model),
        'gls' (Generalized Least Squares),
        'lowess' (Locally Weighted Regression (simple)),
        'loess' (Locally Weighted Regression),
        'mavg' (Moving Average),
        'gpr' (Gaussian Process Regressor)}.
    """

    if mask == None:
        final_mask = np.ones(adata.shape[0]).astype(bool)
    else:
        if mask.shape[0] == adata.shape[0]:
            final_mask = mask
        else:
            logg.error(
                "mask must be a boolean array with the same size as adata.shape[0]."
            )
            return None

    hf.check_available_map(adata)
    fig_width = settings.fig_width
    fig_height = settings.fig_height
    point_size = settings.fig_point_size

    if len(adata.uns["available_map"]) == 0:
        logg.error(f"There is no transition map available yet")

    else:

        if type(selected_fate) == str:
            selected_fate = [selected_fate]

        (
            mega_cluster_list,
            valid_fate_list,
            fate_array_flat,
            sel_index_list,
        ) = hf.analyze_selected_fates(adata.obs["state_info"], selected_fate)
        if len(mega_cluster_list) == 0:
            logg.error("No cells selected. Computation aborted!")
            return adata
        else:
            fate_name = mega_cluster_list[0]
            target_idx = sel_index_list[0]

            x_emb = adata.obsm["X_emb"][:, 0]
            y_emb = adata.obsm["X_emb"][:, 1]
            data_des = adata.uns["data_des"][-1]
            data_path = settings.data_path
            figure_path = settings.figure_path
            file_name = os.path.join(
                data_path, f"{data_des}_fate_trajectory_pseudoTime_{fate_name}.npy"
            )

            traj_name = f"diff_trajectory_{source}_{fate_name}"
            if traj_name not in adata.obs.keys():
                logg.error(
                    f"The target fate trajectory for {fate_name} with {source} have not been inferred yet.\n"
                    "Please infer the trajectory with first with cs.tl.progenitor, \n"
                    "or cs.tl.iterative_differentiation."
                )

            else:
                prob_0 = np.array(adata.obs[traj_name])

                sel_cell_idx = (prob_0 > traj_threshold * np.max(prob_0)) & final_mask
                if np.sum(sel_cell_idx) == 0:
                    raise ValueError("No cells selected!")

                sel_cell_id = np.nonzero(sel_cell_idx)[0]

                if os.path.exists(file_name) and (not compute_new):
                    logg.info("Load pre-computed pseudotime")
                    PseudoTime = np.load(file_name)
                else:

                    from sklearn import manifold

                    data_matrix = adata.obsm["X_pca"][sel_cell_idx]
                    method = SpectralEmbedding(n_components=1, n_neighbors=n_neighbors)
                    PseudoTime = method.fit_transform(data_matrix)
                    np.save(file_name, PseudoTime)
                    # logg.info("Run time:",time.time()-t)

                PseudoTime = PseudoTime - np.min(PseudoTime)
                PseudoTime = (PseudoTime / np.max(PseudoTime)).flatten()

                ## re-order the pseudoTime such that the target fate has the pseudo time 1.
                if invert_PseudoTime:
                    # target_fate_id=np.nonzero(target_idx)[0]
                    # convert_fate_id=hf.converting_id_from_fullSpace_to_subSpace(target_fate_id,sel_cell_id)[0]
                    # if np.mean(PseudoTime[convert_fate_id])<0.5: PseudoTime=1-PseudoTime
                    PseudoTime = 1 - PseudoTime

                # pdb.set_trace()
                if (
                    np.sum((PseudoTime > 0.25) & (PseudoTime < 0.75)) == 0
                ):  # the cell states do not form a contiuum. Plot raw data instead
                    logg.error(
                        "The selected cell states do not form a connected graph. Cannot form a continuum of pseudoTime. Only plot the raw data"
                    )
                    plot_raw_data = True

                ## plot the pseudotime ordering
                fig = plt.figure(figsize=(fig_width * 2, fig_height))
                ax = plt.subplot(1, 2, 1)
                pl_util.customized_embedding(
                    x_emb,
                    y_emb,
                    sel_cell_idx,
                    ax=ax,
                    title="Selected cells",
                    point_size=point_size,
                )
                ax1 = plt.subplot(1, 2, 2)
                pl_util.customized_embedding(
                    x_emb[sel_cell_idx],
                    y_emb[sel_cell_idx],
                    PseudoTime,
                    ax=ax1,
                    title="Pseudotime",
                    point_size=point_size,
                )
                # customized_embedding(x_emb[final_id],y_emb[final_id],PseudoTime,ax=ax1,title='Pseudo time')
                Clb = fig.colorbar(
                    plt.cm.ScalarMappable(cmap=plt.cm.Reds), ax=ax1, label="Pseudotime"
                )
                plt.tight_layout()
                fig.savefig(
                    os.path.join(
                        figure_path,
                        f"{data_des}_fate_trajectory_pseudoTime_{fate_name}.{settings.file_format_figs}",
                    )
                )

                temp_dict = {"PseudoTime": PseudoTime}
                for gene_name in gene_name_list:
                    yy_max = np.percentile(
                        adata.obs_vector(gene_name), gene_exp_percentile
                    )  # global blackground
                    yy = np.array(adata.obs_vector(gene_name)[sel_cell_idx])
                    rescaled_yy = (
                        yy * prob_0[sel_cell_idx] / yy_max
                    )  # rescaled by global background
                    temp_dict[gene_name] = rescaled_yy

                from plotnine import (
                    aes,
                    geom_point,
                    ggplot,
                    labs,
                    stat_smooth,
                    theme_classic,
                )

                data2 = pd.DataFrame(temp_dict)
                data2_melt = pd.melt(
                    data2, id_vars=["PseudoTime"], value_vars=gene_name_list
                )
                gplot = (
                    ggplot(
                        data=data2_melt,
                        mapping=aes(x="PseudoTime", y="value", color="variable"),
                    )
                    + (
                        geom_point()
                        if plot_raw_data
                        else stat_smooth(method=stat_smooth_method)
                    )
                    + theme_classic(base_size=ggplot_font_size)
                    + labs(
                        x="Pseudotime",
                        y="Normalized gene expression",
                        color="Gene name",
                    )
                )

                gplot.save(
                    os.path.join(
                        figure_path,
                        f"{data_des}_fate_trajectory_pseutoTime_gene_expression_{fate_name}.{settings.file_format_figs}",
                    ),
                    width=fig_width,
                    height=fig_height,
                    verbose=False,
                )
                gplot.draw()


def gene_expression_heatmap(
    adata,
    selected_genes=None,
    selected_fates=None,
    rename_fates=None,
    color_bar=True,
    method="relative",
    fig_width=6,
    fig_height=3,
    horizontal="True",
    vmin=None,
    vmax=None,
    figure_index="",
    order_map_x=True,
    order_map_y=True,
    **kwargs,
):
    """
    Plot heatmap of gene expression within given clusters.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
    selected_genes: `list`, optional (default: None)
        A list of selected genes.
    selected_fates: `list`, optional (default: all)
        List of cluster ids consistent with adata.obs['state_info'].
        It allows a nested structure. If so, we merge clusters within
        each sub-list into a mega-fate cluster.
    method: `str`, optional (default: 'relative')
        Method to normalize gene expression. Options: {'relative','zscore',f'relative_{j}'}.
        'relative': given coarse-grained gene expression
        in given clusters, normalize the expression across clusters to be 1;
        'zscore': given coarse-grained gene expression in given clusters, compute its zscore.
        f'relative_{j}': use j-th row for normalization
    rename_fates: `list`, optional (default: None)
        Provide new names in substitution of names in selected_fates.
        For this to be effective, the new name list needs to have names
        in exact correspondence to those in the old list.
    color_bar: `bool`, optional (default: True)
        If true, show the color bar.
    fig_width: `int`, optional (default: 6)
        Figure width.
    fig_height: `int`, optional (default: 3)
        Figure height.
    horizontal: `bool`, optional (default: True)
        Figure orientation.
    vmin: `float`, optional (default: None)
        Minimum value to show.
    vmax: `float`, optional (default: None)
        Maximum value to show.
    figure_index:
        A string for index the figure names.
    order_map_x: `bool`
        Whether to re-order the x coordinate of the matrix or not
    order_map_y: `bool`
        Whether to re-order the y coordinate of the matrix or not

    Returns
    -------
    gene_expression_matrix: `np.array`
    """

    if (method not in ["relative", "zscore"]) and (not method.startswith("relative_")):
        logg.warn("method not in ['relative','zscore']; set it to be 'relative'")
        method = "relative"

    if method.startswith("relative_"):
        # assuming method=f'relative_{j}'
        rela_idx = int(method.split("relative_")[1])
        logg.info(f"Use {rela_idx} row for normalization")

    gene_list = selected_genes
    state_info = np.array(adata.obs["state_info"])
    (
        mega_cluster_list,
        valid_fate_list,
        fate_array_flat,
        sel_index_list,
    ) = hf.analyze_selected_fates(state_info, selected_fates)
    gene_full = np.array(adata.var_names)
    gene_list = np.array(gene_list)
    sel_idx = np.in1d(gene_full, gene_list)
    valid_sel_idx = np.in1d(gene_list, gene_full)

    if np.sum(valid_sel_idx) > 0:
        cleaned_gene_list = gene_list[valid_sel_idx]
        if np.sum(valid_sel_idx) < len(gene_list):
            invalid_gene_list = gene_list[~valid_sel_idx]
            logg.info(f"These are invalid gene names: {invalid_gene_list}")
    else:
        raise ValueError("No valid genes selected.")
    gene_expression_matrix = np.zeros((len(mega_cluster_list), len(cleaned_gene_list)))

    X = adata.X
    resol = 10 ** (-10)

    if method == "zscore":
        logg.hint("Using zscore (range: [-2,2], or [-1,1]")
        color_bar_label = "Z-score"
    else:
        logg.hint("Using relative gene expression. Range [0,1]")
        color_bar_label = "Relative expression"

    for k, temp in enumerate(cleaned_gene_list):
        temp_id = np.nonzero(gene_full == temp)[0][0]
        temp_vector = np.zeros(len(sel_index_list))
        for j, temp_idx in enumerate(sel_index_list):
            temp_vector[j] = np.mean(X[temp_idx, temp_id])

        if method == "zscore":
            z_score = stats.zscore(temp_vector)
            gene_expression_matrix[:, k] = z_score
        elif method == "relative":
            temp_vector = (temp_vector + resol) / (resol + np.sum(temp_vector))
            gene_expression_matrix[:, k] = temp_vector
        else:
            temp_vector = (temp_vector + resol) / (resol + temp_vector[rela_idx])
            gene_expression_matrix[:, k] = temp_vector

    if (rename_fates is None) or (len(rename_fates) != len(mega_cluster_list)):
        rename_fates = mega_cluster_list

    if horizontal:
        ax = pl_util.heatmap(
            gene_expression_matrix,
            x_ticks=cleaned_gene_list,
            y_ticks=rename_fates,
            log_transform=False,
            color_map=plt.cm.coolwarm,
            fig_width=fig_width,
            fig_height=fig_height,
            color_bar=color_bar,
            vmin=vmin,
            vmax=vmax,
            color_bar_label=color_bar_label,
            order_map_x=order_map_x,
            x_tick_style="italic",
            order_map_y=order_map_y,
            **kwargs,
        )
    else:
        ax = pl_util.heatmap(
            gene_expression_matrix.T,
            x_ticks=rename_fates,
            y_ticks=cleaned_gene_list,
            log_transform=False,
            color_map=plt.cm.coolwarm,
            fig_width=fig_height,
            fig_height=fig_width,
            color_bar=color_bar,
            vmin=vmin,
            vmax=vmax,
            color_bar_label=color_bar_label,
            order_map_x=order_map_x,
            y_tick_style="italic",
            order_map_y=order_map_y,
            **kwargs,
        )

    plt.tight_layout()
    if figure_index != "":
        figure_index == f"_{figure_index}"

    data_des = adata.uns["data_des"][-1]
    plt.savefig(
        os.path.join(
            settings.figure_path,
            f"{data_des}_gene_expression_matrix{figure_index}.{settings.file_format_figs}",
        )
    )
    return gene_expression_matrix


def gene_expression_on_manifold(
    adata, selected_genes, savefig=False, selected_times=None, color_bar=False
):
    """
    Plot gene expression on the state manifold.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    selected_genes: `list` or 'str'
        List of genes to plot.
    savefig: `bool`, optional (default: False)
        Save the figure.
    selected_times: `list`, optional (default: all)
        A list of time points to further restrict the cell states to plot.
        If map_backward=True, plot initial states that are among these time points;
        otherwise, show later states that are among these time points.
    color_bar: `bool`, (default: False)
        If True, plot the color bar.
    """

    selected_genes = list(selected_genes)
    fig_width = settings.fig_width
    fig_height = settings.fig_height
    point_size = settings.fig_point_size
    if color_bar:
        fig_width = fig_width + 0.5

    if type(selected_genes) == str:
        selected_genes = [selected_genes]

    x_emb = adata.obsm["X_emb"][:, 0]
    y_emb = adata.obsm["X_emb"][:, 1]
    figure_path = settings.figure_path

    time_info = np.array(adata.obs["time_info"])
    if selected_times is not None:
        sp_idx = np.zeros(adata.shape[0], dtype=bool)
        for xx in selected_times:
            sp_id_temp = np.nonzero(time_info == xx)[0]
            sp_idx[sp_id_temp] = True
    else:
        sp_idx = np.ones(adata.shape[0], dtype=bool)

    for j, g in enumerate(selected_genes):
        vector = adata[:, g].X.A.flatten()
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax = plt.subplot(1, 1, 1)
        pl_util.customized_embedding(
            x_emb[sp_idx],
            y_emb[sp_idx],
            vector[sp_idx],
            title=g,
            ax=ax,
            col_range=(0, 99.8),
            color_bar=color_bar,
            point_size=point_size,
            color_bar_label="Normalized expression",
        )
        plt.title(g, style="italic")

        plt.tight_layout()
        if savefig:
            plt.savefig(
                os.path.join(
                    figure_path,
                    f"gene_expression_{selected_genes[j]}.{settings.file_format_figs}",
                )
            )



#初始化函数，将kailin转至工作目录。如果此前初始化过，那么在再次运行def kl_initialize(0)时，
#则拒绝初始化，避免套娃。运行def kl_initialize(1)时，强制重新初始化。

print('kailin has been initialized')
def kl_initialize(reset):

    import os
    import LittleSnowFox
    global empty_key
    global parent_directory_origin
    
    
    #初始化的具体内容
    def first_time():
        global empty_key
        global parent_directory_origin
        empty_key = 0
        print('set empty_key = ' + str(empty_key))

        # 获取 kailin 模块所在的目录
        kailin_directory = os.path.dirname(os.path.abspath(LittleSnowFox.__file__))
        print(f"The directory of the kailin module is: {kailin_directory}")
        parent_directory_origin = kailin_directory
        print('work address = ' + str(parent_directory_origin))
        print()
        print()
        print('Note--If you want use your data, please put your data.h5ad in: ' + parent_directory_origin + '\database')
        print()
        print()



        

    #如果此前初始化过，则拒绝初始化，避免套娃
    if 'empty_key' in locals() or 'empty_key' in globals():
        print("You have initialized before, the current parameter is:")
        #print('empty_key = ' + str(empty_key))
        print('work address = ' + str(parent_directory_origin))
        print()
        print()
        print('Note--If you want use your data, please put your data.h5ad in: ' + parent_directory_origin + '\database')
        print()
        print()

        #重设模式，初始化参数重设到当前工作区
        if reset == 1:
            print("reset mode")
            first_time()

    #没有初始化过，进行初始化
    else:
        print("empty_key does not exist")
        first_time()

    return empty_key
    return parent_directory_origin




import os
import time
from logging import raiseExceptions

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from scipy.cluster import hierarchy

# from plotnine import *
from sklearn.manifold import SpectralEmbedding

from cospar.tool import _utils as tl_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings


def differential_genes(
    adata,
    cell_group_A=None,
    cell_group_B=None,
    FDR_cutoff=0.05,
    sort_by="ratio",
    min_frac_expr=0.05,
    pseudocount=1,
):
    """
    Perform differential gene expression analysis and plot top DGE genes.

    We use Wilcoxon rank-sum test to calculate P values, followed by
    Benjamini-Hochberg correction.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Need to contain gene expression matrix.
    cell_group_A: `np.array`, optional (default: None)
        A boolean array of the size adata.shape[0] for defining population A.
        If not specified, we set it to be adata.obs['cell_group_A'].
    cell_group_B: `np.array`, optional (default: None)
        A boolean array of the size adata.shape[0] for defining population B.
        If not specified, we set it to be adata.obs['cell_group_A'].
    FDR_cutoff: `float`, optional (default: 0.05)
        Cut off for the corrected Pvalue of each gene. Only genes below this
        cutoff will be shown.
    sort_by: `float`, optional (default: 'ratio')
        The key to sort the differentially expressed genes. The key can be: 'ratio' or 'Qvalue'.
    min_frac_expr: `float`, optional (default: 0.05)
        Minimum expression fraction among selected states for a
        gene to be considered for DGE analysis.
    pseudocount: `int`, optional (default: 1)
        pseudo count for taking the gene expression ratio between the two groups

    Returns
    -------
    diff_gene_A: `pd.DataFrame`
        Genes differentially expressed in cell state group A, ranked
        by the ratio of mean expressions between
        the two groups, with the top being more differentially expressed.
    diff_gene_B: `pd.DataFrame`
        Genes differentially expressed in cell state group B, ranked
        by the ratio of mean expressions between
        the two groups, with the top being more differentially expressed.
    """

    diff_gene_A = []
    diff_gene_B = []

    if sort_by not in ["ratio", "Qvalue"]:
        raise ValueError(f"sort_by must be among {['ratio','Qvalue']}")

    state_info = np.array(adata.obs["state_info"])
    inputs = [cell_group_A, cell_group_B]
    selections = []
    for cell_group_X in inputs:
        if type(cell_group_X) is str:
            if cell_group_X in list(set(state_info)):
                group_idx = state_info == cell_group_X
            else:
                raise ValueError(
                    "cell_group_A (or B) should be either a cluster name among adata.obs['state_info'] or a boolean array of size adata.shape[0]."
                )
        else:
            group_idx = np.array(cell_group_X).astype("bool")

        selections.append(group_idx)

    if (np.sum(selections[0]) == 0) or (np.sum(selections[1]) == 0):
        raise ValueError("Group A or B has zero selected cell states.")

    else:

        dge = hf.get_dge_SW(
            adata,
            selections[0],
            selections[1],
            min_frac_expr=min_frac_expr,
            pseudocount=pseudocount,
        )

        dge = dge.sort_values(by=sort_by, ascending=False)
        diff_gene_A_0 = dge
        diff_gene_A = diff_gene_A_0[(dge["Qvalue"] < FDR_cutoff) & (dge["ratio"] > 0)]
        diff_gene_A = diff_gene_A.reset_index()

        dge = dge.sort_values(by=sort_by, ascending=True)
        diff_gene_B_0 = dge
        diff_gene_B = diff_gene_B_0[(dge["Qvalue"] < FDR_cutoff) & (dge["ratio"] < 0)]
        diff_gene_B = diff_gene_B.reset_index()

    return diff_gene_A, diff_gene_B


def identify_TF_and_surface_marker(
    gene_list,
    species="mouse",
    go_term_keywards=[
        "cell surface",
        "cell cycle",
        "regulation of transcription",
        "DNA-binding transcription factor activity",
        "regulation of transcription by RNA polymerase II",
    ],
):
    """
    From an input gene list, return the go term and annotation for each gene,
    and further select the genes identified as TF or cell surface protein

    Returns
    ------
        results:
            Full annotation for each gene
        df_anno
            Only include genes identified as TF or cell surface protein
    """

    if species not in ["mouse", "human"]:
        raise ValueError("species must be either mouse or human")
    else:
        if species == "mouse":
            dataset = "mmusculus_gene_ensembl"
        elif species == "human":
            dataset = "hsapiens_gene_ensembl"

    from gseapy.parser import Biomart

    bm = Biomart()
    ## view validated marts
    marts = bm.get_marts()
    ## view validated dataset
    datasets = bm.get_datasets(mart="ENSEMBL_MART_ENSEMBL")
    ## view validated attributes
    attrs = bm.get_attributes(
        dataset=dataset
    )  # hsapiens_gene_ensembl: Human genes (GRCh38.p13);  mmusculus_gene_ensembl for 'Mouse genes (GRCm39)'
    ## view validated filters
    filters = bm.get_filters(dataset=dataset)  # Gene Name(s) [e.g. MT-TF]
    ## query results

    results = bm.query(
        dataset=dataset,
        attributes=[
            "ensembl_gene_id",
            "external_gene_name",
            "namespace_1003",
            "name_1006",
        ],
        filters={"external_gene_name": gene_list},
    )
    results = results.dropna()
    df_list = []
    for term in go_term_keywards:
        tmp_genes = list(
            set(
                results[results["name_1006"].apply(lambda x: term == x)][
                    "external_gene_name"
                ]
            )
        )
        df_tmp = pd.DataFrame({"gene": tmp_genes})
        df_tmp["annotation"] = term
        df_list.append(df_tmp)
    df_anno = pd.concat(df_list, ignore_index=True)
    return results, df_anno



from textwrap import dedent


def _doc_params(**kwds):
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """

    def dec(obj):
        obj.__orig_doc__ = obj.__doc__
        obj.__doc__ = dedent(obj.__doc__).format_map(kwds)
        return obj

    return dec


selected_fates = """\
selected_fates: `list`
    List of cluster ids consistent with adata.obs['state_info'].
    It allows a nested structure. If so, we merge clusters within
    each sub-list into a mega-fate cluster.\
"""

map_source = """\
source: `str`
    The transition map to be used for plotting: {'transition_map',
    'intraclone_transition_map',...}. The actual available
    map depends on adata itself, which can be accessed at adata.uns['available_map']\
"""

map_backward = """\
map_backward: `bool`, optional (default: True)
    If `map_backward=True`, show fate properties of initial cell states :math:`i`;
    otherwise, show progenitor properties of later cell states :math:`j`.
    This is used for building the fate map :math:`P_i(\mathcal{C})`. See :func:`.fate_map`.\
"""

fate_method = """\
method: `str`, optional (default: 'norm-sum')
    Method to obtain the fate probability map :math:`P_i(\mathcal{C})` towards a set
    of states annotated with fate :math:`\mathcal{C}`. Available options:
    {'sum', 'norm-sum'}. See :func:`.fate_map`.\
"""

sum_fate_prob_thresh = """\
sum_fate_prob_thresh: `float`, optional (default: 0.05)
    The fate bias of a state is plotted only when it has a cumulative fate
    probability to the combined cluster (A+B) larger than this threshold,
    i.e., P(i->A)+P(i+>B) >  sum_fate_prob_thresh.\
"""

selected_times = """\
selected_times: `list`, optional (default: all)
    A list of time points to further restrict the cell states to plot.
    The default choice is not to constrain the cell states to show.\
"""

all_source = """\
source: `str`
    Choices: {'X_clone', 'transition_map',
    'intraclone_transition_map',...}. If set to be 'clone', use only the clonal
    information. If set to be any of the precomputed transition map, use the
    transition map to compute the fate coupling. The actual available
    map depends on adata itself, which can be accessed at adata.uns['available_map']\
"""


rename_fates = """\
rename_fates: `list`, optional (default: None)
    Provide new names in substitution of names in selected_fates.
    For this to be effective, the new name list needs to have names
    in exact correspondence to those in the old list.\
"""


background = """\
background: `bool`, optional (default: True)
    If true, plot all cell states (t1+t2) in grey as the background.\
"""

show_histogram = """\
show_histogram: `bool`, optional (default: False)
    If true, show the distribution of inferred fate probability.\
"""

plot_target_state = """\
plot_target_state: `bool`, optional (default: True)
    If true, highlight the target clusters as defined in selected_fates.\
"""

color_bar = """\
color_bar: `bool`, optional (default: True)
    plot the color bar if True.\
"""

auto_color_scale = """\
auto_color_scale:
    True: automatically rescale the color range to match the value range.\
"""

target_transparency = """\
target_transparency: `float`, optional (default: 0.2)
    It controls the transparency of the plotted target cell states,
    for visual effect. Range: [0,1].\
"""

figure_index = """\
figure_index: `str`, optional (default: '')
    String index for annotate filename for saved figures. Used to distinuigh plots from different conditions.\
"""

mask = """\
mask: `np.array`, optional (default: None)
    A boolean array for available cell states. It should has the length as adata.shape[0].
    Especially useful to constrain the states to show fate bias.\
"""

color_map = """\
color_map:
    The color map (a matplotlib.pyplot.cm object) to visualize the result.\
"""



# -*- coding: UTF-8 -*-
import matplotlib
matplotlib.use('Agg')
import os
from xml.dom.expatbuilder import InternalSubsetExtractor
import random
import anndata
import numpy as np
import scanpy as sc
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from scipy.sparse import find
from scipy.sparse import csr_matrix



def plot_Colormap(X_clone, T_data, X1, X2, node_size, save_path, 
                  with_labels=False, draw_edges=True, 
                  specified_tree_index=None, specified_node_name=None):

    #X_clone是读取h5ad的anndata里的adata.obsm['X_clone']
    #C_data是读取聚类数据csv，也就是pd.read_csv('combined_monocle2.csv')而combined_monocle2.csv是matlab里对count_result_rank直接使用writetable存的文件
    #T_data是读取伪时间数据csv，也就是R语言写的monocle2读取的伪时间存成的pseudotime_data_monocle2.csv
    #X1,X2是每个点的坐标，分别是adata.obsm['X_emb']的第一列和第二列
    #node_size是所作图像的点的大小
    #save_path是所作图像的存储路径

    # 染色方案
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_gray_blue", ["#17324b", "#61ade8"])
    
    index_map = {name: idx for idx, name in enumerate(T_data['cell_name'])}
    rev_index_map = {idx: name for idx, name in enumerate(T_data['cell_name'])}
    # 创建伪时间数组
    T = csv_data['pseudotime']
    lenT = len(T)
    # 创建有向图
    G = nx.DiGraph()

    # 从稀疏矩阵中找出非零元素的位置
    rows, cols, _ = find(X_clone)  

    # 将稀疏矩阵逐点提取，以细胞索引为点，逐个添加有向边
    for row, col in zip(rows, cols):
        if row < lenT and col < lenT:
            if T[row] < T[col]:
                G.add_edge(row, col)
            else:
                G.add_edge(col, row)

    # 识别所有弱连通分量
    weak_components = list(nx.weakly_connected_components(G))
    forest = []

    # 遍历每个弱连通分量
    for component in weak_components:
        subgraph = G.subgraph(component).copy()
        roots = [node for node in subgraph if subgraph.in_degree(node) == 0]
        if not roots:
            roots = [next(iter(component))]

        for root in roots:
            tree = nx.dfs_tree(subgraph, source=root)
            forest.append(tree)

    # 确保保存路径存在
    os.makedirs(save_path, exist_ok=True)

    # 根据指定的参数绘制图像

    trees_to_plot = []
    original_indices = []

    if specified_tree_index is not None:
        if isinstance(specified_tree_index, int):
            specified_tree_index = [specified_tree_index]
        trees_to_plot = [forest[i] for i in specified_tree_index if i < len(forest)]
        original_indices = [i for i in specified_tree_index if i < len(forest)]
    elif specified_node_name is not None:
        if isinstance(specified_node_name, int):
            specified_node_name = [specified_node_name]
        node_indices = [index_map.get(name) for name in specified_node_name if name in index_map]
        for index, tree in enumerate(forest):
            if any(node in tree.nodes for node in node_indices):
                trees_to_plot.append(tree)
                original_indices.append(index)
    else:
        trees_to_plot = forest
        original_indices = list(range(len(forest)))

    all_trees = len(trees_to_plot)
    # 绘制指定的树
    for i, (index, tree) in enumerate(zip(original_indices, trees_to_plot), start=1):
        print(f"Plotting... ( {i} / {all_trees} )", end='\r')  # 输出当前进度
        plt.figure(figsize=(16, 9))
        plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)
        pos = {node: (X1[node], X2[node]) for node in tree.nodes() if node < lenT}
        node_colors = [T[node] for node in tree.nodes() if node < lenT]
        labels = {node: rev_index_map[node] for node in tree.nodes() if node in rev_index_map}
        if draw_edges:
            nx.draw(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap,
                   labels=labels if with_labels else None, edge_color='gray', node_size=node_size, font_size=8)
        else:
            nx.draw_networkx_nodes(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap, node_size=node_size)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=min(node_colors), vmax=max(node_colors)))
        plt.colorbar(sm, orientation='vertical', label='Pseudotime')
        plt.title(f"UMAP Plot with Pseudotime Coloring - Tree {index}")
        plt.xlabel('UMAP 1')
        plt.ylabel('UMAP 2')
        plt.savefig(os.path.join(save_path, f'Tree_{index}.png'))
        plt.close()




def plot_Clusters(C_data, T_data, X1, X2, node_size, save_path, Color="#618a03",
                 specified_cluster_index=None,specified_node_name=None):

    #X_clone是读取h5ad的anndata里的adata.obsm['X_clone']
    
    #C_data是读取聚类数据csv，也就是pd.read_csv('combined_monocle2.csv')
    #而combined_monocle2.csv是matlab里对count_result_rank直接使用writetable存的文件
    
    #T_data是读取伪时间数据csv，也就是R语言写的monocle2读取的伪时间存成的pseudotime_data_monocle2.csv
    #X1,X2是每个点的坐标，分别是adata.obsm['X_emb']的第一列和第二列
    #node_size是所作图像的点的大小
    #save_path是所作图像的存储路径

    # 根据MM_new_rank分组细胞名到clusters
    clusters = C_data.groupby('MM_new_rank')['Var3'].apply(set).to_dict()
    # 创建从细胞名到索引的映射
    index_map = {name: idx for idx, name in enumerate(T_data)}

    plt.figure(figsize=(16, 9))
    # 绘制背景点
    plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)

    # 处理输入参数，允许单个或多个输入
    if isinstance(specified_cluster_index, int):
        specified_cluster_index = [specified_cluster_index]
    if isinstance(specified_node_name, int):
        specified_node_name = [specified_node_name]


    # 处理颜色输入，允许单个颜色或多颜色输入
    if isinstance(Color, str):
        Color = [Color]

     # 确定需要绘制的簇
    if specified_node_name:
        specified_cluster_index = [idx for idx, cells in clusters.items()
                                   if any(node in cells for node in specified_node_name)]
        filename_suffix = f'_node{",".join(map(str, specified_node_name))}'
    elif specified_cluster_index:
        filename_suffix = f'_cell{",".join(map(str, specified_cluster_index))}'
    else:
        specified_cluster_index = list(clusters.keys())
        filename_suffix = '_all'

    # 绘制指定的簇
    for i, idx in enumerate(specified_cluster_index):
        if idx in clusters:
            node_indices = [index_map[node] for node in clusters[idx] if node in index_map]
            color = Color[i % len(Color)] if i < len(Color) else f"#{random.randint(0, 0xFFFFFF):06x}"
            plt.scatter(X1[node_indices], X2[node_indices], c=color, s=node_size, label=f'Cluster {idx}')
    # 在图中特别标出 specified_node_name 对应的节点
    if specified_node_name:
        special_indices = [index_map[node] for node in specified_node_name if node in index_map]
        plt.scatter(X1[special_indices], X2[special_indices], c='red', s=node_size+2, linewidths=1)
        for idx in special_indices:
            plt.text(X1[idx], X2[idx], f'{T_data[idx]}', fontsize=9, color='black', ha='right')
    # 添加图例
    plt.legend(title='Clusters')
    plt.title("UMAP Plot with Cluster Coloring")
    plt.xlabel('UMAP 1')
    plt.ylabel('UMAP 2')
    plt.savefig(f'{save_path}/clusters_plot{filename_suffix}.png')
    plt.close()

'''
#def get_row_data(sparse_matrix, row_index):
#    if row_index < sparse_matrix.shape[0]:  # 确保行索引有效
#        row_start = sparse_matrix.indptr[row_index]
#        row_end = sparse_matrix.indptr[row_index + 1]
#        columns = sparse_matrix.indices[row_start:row_end]
#        values = sparse_matrix.data[row_start:row_end]
#        return columns, values
#    else:
#        return None, None

# 路径设置
file_path = 'Hematopoiesis_progenitor.h5ad'
csv_path =  'combined_data_monocle3.csv'
cluster_path = 'combined_monocle2.csv'

# 读取H5AD
adata = anndata.read_h5ad(file_path)
# 读取CSV
csv_data = pd.read_csv(csv_path)
cluster_data = pd.read_csv(cluster_path)

# 获取X_clone矩阵
X_clone = adata.obsm['X_clone']

#读取umap坐标 
umap_df = pd.DataFrame(adata.obsm['X_emb'], columns=['X1', 'X2'])
X1 = umap_df['X1']
X2 = umap_df['X2']

node_size=8

save_path='G:/ASUMDownloads/dataplot/'
#indexA = 2097
#indexB = 392
## 获取第2097行和第392行的数据
#columns_A, values_A = get_row_data(X_clone, indexA)
#columns_B, values_B = get_row_data(X_clone, indexB)

## 输出结果
#print(f"Columns and values in row {indexA}:")
#print("Columns:", columns_A)
#print("Values:", values_A)

#print(f"Columns and values in row {indexB}:")
#print("Columns:", columns_B)
#print("Values:", values_B)
# 可以在这里入26个科研常用的颜色
#colors = [
#    "#e6194b", "#3cb44b", "#ffe119", "#4363d8", "#f58231", "#911eb4", "#46f0f0",
#    "#f032e6", "#bcf60c", "#fabebe", "#008080", "#e6beff", "#9a6324", "#fffac8",
#    "#800000", "#aaffc3", "#808000", "#ffd8b1", "#000075", "#808080",
#    "#ffffff", "#000000", "#fabed4", "#ffd700", "#dcbeff", "#f4901b"
#]

# 示例调用
#plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path)
#plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path,specified_node_name=9429)
plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path,specified_cluster_index=[1,2,10,12,13,19,20,21,22,23,24,25,26 ])

#for i in range(1,27):
#    plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path,specified_cluster_index=i)

#plot_Clusters(X_clone,cluster_data, csv_data, X1, X2, node_size, save_path, Color="#61ade8",specified_node_name=37803)
# 稀疏矩阵 伪时间 X.Y的UMAP,节点大小,存储路径,是否显示标签,是否显示边
#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_tree_index=232)

#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_tree_index=[233,234])

#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=False,draw_edges=False,specified_node_name=32689)

#plot_Colormap(X_clone,csv_data,X1,X2,node_size,save_path,with_labels=True,draw_edges=True,specified_node_name=[5314,4465])

#plt.ion()
'''


import os
import time

import numpy as np
import scanpy as sc
import scipy.sparse as ssp
from matplotlib import pyplot as plt
from tqdm import tqdm

from . import hf, settings, tl


def sigma(x, diff_sigma):
    #    return 0.1*np.sin(0.1*x)+0.1
    return diff_sigma


# growth is not used below
def growth(x):
    #    return 0.1*np.cos(0.2*x)+1
    return 1


def progression(x, dL):
    #    return 0.1*np.sin(0.1*x)+1
    return dL


def kernel(x, y_vec, smooth_sigma):
    temp = []
    for y in y_vec:
        d = abs(x - y)

        temp.append(np.exp(-((d) ** 2) / (2 * (smooth_sigma**2))))

    norm_Prob = np.array(temp) / np.sum(temp)

    return norm_Prob


def kernel_matrix(x_vec, smooth_sigma):
    matrix = np.zeros((len(x_vec), len(x_vec)))
    for j, x in enumerate(x_vec):
        matrix[j] = kernel(x, x_vec, smooth_sigma)

    return matrix


def kernel_matrix_v1(smooth_sigma, shortest_distance):
    shortest_distance = np.array(shortest_distance)
    matrix_temp = np.exp(-((shortest_distance) ** 2) / (2 * (smooth_sigma**2)))
    norm_kernel = np.zeros(matrix_temp.shape)
    for j in range(matrix_temp.shape[0]):
        norm_kernel[j] = matrix_temp[j] / np.sum(matrix_temp[j])

    return norm_kernel


def shortest_path_distance(x_vec, bif_x=0, mode="1d"):
    N = len(x_vec)
    D = np.zeros((N, N))  # distance matrix

    for i, x in enumerate(x_vec):
        for j, y in enumerate(x_vec):
            if mode == "1d":
                d = abs(x - y)
            else:
                # this is the bifurcation model
                if x[1] == y[1]:  # belonging to the same branch
                    d = abs(x[0] - y[0])
                else:
                    d = abs(x[0] - bif_x) + abs(y[0] - bif_x)

            D[i, j] = d

    return D


def transition_prob(x, y_vec, diff_sigma, dL):
    temp = []
    for y in y_vec:
        temp.append(
            np.exp(
                -((x - y + progression(x, dL)) ** 2) / (2 * (sigma(x, diff_sigma) ** 2))
            )
        )

    norm_Prob = np.array(temp) / np.sum(temp)

    return norm_Prob


def simulate_cell_position_next_time_point(x_t0, parameter, mode="1d"):
    L = parameter["L"]
    dL = parameter["dL"]
    diff_sigma = parameter["diff_sigma"]
    progeny_N = parameter["progeny_N"]
    dx = parameter["dx"]
    lattice = parameter["lattice"]
    bifurcation = parameter["bifurcation"]

    x_next = []
    id_next = []

    if mode == "1d":
        # 1 d problem
        for j, x in enumerate(x_t0):
            ## this method is much faster
            if x < L - (1.5 * progression(x, dL) + 5 * sigma(x, diff_sigma)):
                bb = np.random.normal(
                    x + progression(x, dL), sigma(x, diff_sigma), progeny_N
                )
                x_next = x_next + list(bb)
                id_next = id_next + list(np.floor(bb / dx).astype(int))
            else:
                ## this method is much slower
                for k in range(progeny_N):
                    cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, dL))
                    # pick the first event that is larger than the random variable
                    new_id = np.nonzero(cum_Prob > np.random.rand())[0][0].astype(int)
                    id_next.append(new_id)
                    x_next.append(lattice[new_id])
    else:

        for j, X in enumerate(x_t0):
            x = X[0]
            ## this method is much faster
            if x < L - (1.5 * progression(x, dL) + 5 * sigma(x, diff_sigma)):
                bb = np.random.normal(
                    x + progression(x, dL), sigma(x, diff_sigma), progeny_N
                )
                id_next = id_next + list(np.floor(bb / dx).astype(int))

            else:
                ## this method is much slower
                bb = []
                for k in range(progeny_N):
                    cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, dL))
                    # pick the first event that is larger than the random variable
                    new_id = np.nonzero(cum_Prob > np.random.rand())[0][0].astype(int)
                    id_next.append(new_id)
                    bb.append(lattice[new_id])

            ## update cell state
            if x > bifurcation:
                x_next = x_next + [[y, X[1]] for y in bb]
            else:
                temp_next = []
                for y in bb:
                    if y < bifurcation:
                        temp_next.append([y, -1])
                    else:
                        fate = np.random.randint(2)
                        temp_next.append([y, fate])
                x_next = x_next + temp_next

    return x_next, id_next


def simulate_cell_position_next_time_point_continuous_barcoding(
    x_t0, parameter, mode="1d", barcoding_rate=0.1
):
    L = parameter["L"]
    dL = parameter["dL"]
    diff_sigma = parameter["diff_sigma"]
    progeny_N = parameter["progeny_N"]
    dx = parameter["dx"]
    lattice = parameter["lattice"]
    bifurcation = parameter["bifurcation"]

    x_next = []
    id_next = []

    if mode == "1d":
        # 1 d problem
        for j, X in enumerate(x_t0):  #
            ## this method is much faster
            x = X[0]  # X[0], location; X[1], barcode information
            if x < L - (1.5 * progression(x, dL) + 5 * sigma(x, diff_sigma)):
                bb = np.random.normal(
                    x + progression(x, dL), sigma(x, diff_sigma), progeny_N
                )
                x_next_temp = []
                for y in bb:
                    barcode_list = X[1]
                    if np.random.rand() < barcoding_rate:
                        barcode_list.append(np.random.randint(1000000))

                    x_next_temp.append([y, barcode_list])
                x_next = x_next + x_next_temp
                id_next = id_next + list(np.floor(bb / dx).astype(int))
            else:
                ## this method is much slower
                for k in range(progeny_N):
                    cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, dL))
                    # pick the first event that is larger than the random variable
                    new_id = np.nonzero(cum_Prob > np.random.rand())[0][0].astype(int)
                    id_next.append(new_id)
                    bb = lattice[new_id]
                    x_next_temp = []
                    for y in bb:
                        barcode_list = X[1]
                        if np.random.rand() < barcoding_rate:
                            barcode_list.append(np.random.randint(1000000))

                        x_next_temp.append([y, barcode_list])
                    x_next = x_next + x_next_temp

    else:

        for j, X in enumerate(x_t0):
            x = X[0]  # X[0], location; X[1], fate choice; X[2], barcode information
            ## this method is much faster
            if x < L - (1.5 * progression(x, dL) + 5 * sigma(x, diff_sigma)):
                bb = np.random.normal(
                    x + progression(x, dL), sigma(x, diff_sigma), progeny_N
                )
                id_next = id_next + list(np.floor(bb / dx).astype(int))

            else:
                ## this method is much slower
                bb = []
                for k in range(progeny_N):
                    cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, dL))
                    # pick the first event that is larger than the random variable
                    new_id = np.nonzero(cum_Prob > np.random.rand())[0][0].astype(int)
                    id_next.append(new_id)
                    bb.append(lattice[new_id])

            ##  fate choice
            fate_choice_list = []
            if x > bifurcation:
                fate_choice_list = [X[1] for y in bb]
            else:
                for y in bb:
                    if y < bifurcation:
                        fate_choice_list.append(-1)
                    else:
                        fate = np.random.randint(2)
                        fate_choice_list.append(fate)

            ## update barcode
            barcode_list_array = []
            for y in bb:
                barcode_list = X[2]
                if np.random.rand() < barcoding_rate:
                    barcode_list.append(np.random.randint(1000000))

                barcode_list_array.append(barcode_list.copy())

            for j in range(len(bb)):
                # pdb.set_trace()
                x_next.append([bb[j], fate_choice_list[j], barcode_list_array[j]])

    return x_next, id_next


def bifurcation_model(
    progeny_N=2,
    t1=5,
    p1=0.5,
    p2=1,
    M=50,
    L=10,
    diff_sigma=0.5,
    repeat_N=1,
    dL=1,
    no_loss=False,
    always_simulate_data=0,
):
    """
    Simulate bifurcation corrupted with clonal dispersion (See Fig. 3e)

    Parameters
    ----------
    progeny_N:
        Fold change of clone size after each generation. 2 means that it will double its clone size after one generation.
    t1:
        Initital sampling time point. Unit: cell cycle. By default
        t2=t1+1.
    p1:
        Probability to sample cells at t1
    p2:
        Probability to sample cells at t2
    M:
        Total number of clones to simulate
    L:
        Total length of the 1-d differentiation manifold
    diff_sigma:
        Differentiation noise
    dL:
        Step size of differentiation for one generation
    no_loss:
        whether the sampling kills the cell or not. Default: sampling kills cells.
        x_t0:
        List of inititial cell locations at t0 along a 1-d line with length L
    always_simulate_data:
        Simulate new data (do not load pre-saved datasets)

    Returns
    -------
    adata:
        An adata object with clonal matrix, time info etc. Ready to be plugged into CoSpar.
    """

    if always_simulate_data:
        file_name = "simulate_data"
    else:
        file_name = f"{settings.data_path}/simulated_clonal_data_bifurcation_M{M}_progeny{progeny_N}_L{L}_dL{dL}_diffSigma{diff_sigma}_t1{t1}_p1{p1}_p2{p2}_simuV{repeat_N}_noloss{no_loss}"
    if os.path.exists(file_name + "_clonal_annot.npz"):
        print("Load existing data")
        clone_annot = ssp.load_npz(file_name + "_clonal_annot.npz")
        simu_data = np.load(file_name + "_others.npz", allow_pickle=True)
        final_coordinates = simu_data["final_coordinates"]
        time_info = simu_data["time_info"]
        bifurcation = 0.5 * L
    #             n=M*(progeny_N**(t1+1))*20 # resolution of grid for sampling
    #             dx=L/n
    else:
        print("Generate new data")
        t = time.time()
        ############################################ simulate the clonal data for each clone
        n = M * (progeny_N ** (t1 + 1)) * 20  # resolution of grid for sampling
        dx = L / n
        lattice = np.linspace(0, L, n)
        t0_id = np.sort(
            np.random.choice(n, M, replace=False)
        )  # random sample of indices
        x_t0_array = np.array(lattice[t0_id])  # position of initial barcoding
        # x_t0_array=np.array(range(20))
        x_t0_array = x_t0_array[x_t0_array < (L - t1 * dL)]
        # print(x_t0_array)

        bifurcation = 0.5 * L

        parameter = {}
        parameter["L"] = L
        parameter["dL"] = dL
        parameter["diff_sigma"] = diff_sigma
        parameter["progeny_N"] = progeny_N
        parameter["dx"] = dx
        parameter["lattice"] = lattice
        parameter["bifurcation"] = bifurcation

        final_coordinates = []
        time_info = []
        clone_id = []
        temp_clone_id = 0
        previouse_cell_N = 0
        # x_t0_array=[0.5,1,3,24,40] # initially barcoded cell positions

        ## simulate the multi-generational drift
        for m31 in tqdm(range(len(x_t0_array))):
            x_t0 = x_t0_array[m31]
            # print("Current clone number:", m31)
            x_next = []
            if x_t0 < bifurcation:
                x_next.append([x_t0, -1])
            else:
                fate = np.random.randint(2)
                x_next.append([x_t0, fate])

            ## 1d simulation
            for j in range(t1):
                x_next, id_next = simulate_cell_position_next_time_point(
                    x_next, parameter, mode="2d"
                )

            ### sampling: t1
            sel_idx = np.random.rand(len(x_next)) < p1
            sel_id = list(np.nonzero(sel_idx)[0].astype(int))
            unsel_id = list(np.nonzero(~sel_idx)[0].astype(int))
            final_coordinates = final_coordinates + list(np.array(x_next)[sel_id])
            time_info = time_info + [1 for j in range(len(sel_id))]

            if no_loss:  # whether the sampling kills the cell or not
                new_x = np.array(x_next)[sel_id]  # remaining cells
            else:
                new_x = np.array(x_next)[unsel_id]  # remaining cells

            # new_x=np.array(x_next)[unsel_id]# remaining cells
            x_next, id_next = simulate_cell_position_next_time_point(
                new_x, parameter, mode="2d"
            )
            ### sampling: t2
            sel_idx = np.random.rand(len(x_next)) < p2
            sel_id = list(np.nonzero(sel_idx)[0].astype(int))
            final_coordinates = final_coordinates + list(np.array(x_next)[sel_id])
            time_info = time_info + [2 for j in range(len(sel_id))]

            clone_id = clone_id + [
                temp_clone_id for j in range(len(time_info) - previouse_cell_N)
            ]
            previouse_cell_N = len(time_info)
            temp_clone_id = temp_clone_id + 1

        ### Generate clonal data matrix
        clone_annot = np.zeros((len(time_info), len(x_t0_array)))
        # ini_clone_id=0
        # jump=np.diff(time_info)
        # clone_id=np.zeros(len(time_info))
        for j in range(len(time_info)):
            clone_annot[j, clone_id[j]] = 1
            # clone_id[j]=ini_clone_id

        clone_annot = ssp.csr_matrix(clone_annot)
        if not always_simulate_data:
            ssp.save_npz(file_name + "_clonal_annot.npz", clone_annot)
            np.savez(
                file_name + "_others.npz",
                final_coordinates=final_coordinates,
                time_info=time_info,
            )
        print("Time elapsed for generating clonal data: ", time.time() - t)

    ## transformation to 50-d
    UMAP_noise = 0.2  #
    final_coordinates = np.array(final_coordinates)
    state_annote_0 = final_coordinates[:, 1]
    x_caleb = np.zeros(len(final_coordinates))
    y_caleb = np.zeros(len(final_coordinates))
    idx = state_annote_0 == -1
    x_caleb[idx] = final_coordinates[idx, 0] - bifurcation
    y_caleb[idx] = 0 * final_coordinates[idx, 0]
    idx = state_annote_0 == 0
    x_caleb[idx] = 0.5 * (final_coordinates[idx, 0] - bifurcation)
    y_caleb[idx] = 0.5 * (final_coordinates[idx, 0] - bifurcation)
    idx = state_annote_0 == 1
    x_caleb[idx] = 0.5 * (final_coordinates[idx, 0] - bifurcation)
    y_caleb[idx] = -0.5 * (final_coordinates[idx, 0] - bifurcation)

    all_coord = np.zeros((len(x_caleb), 50))
    all_coord[:, 0] = x_caleb
    all_coord[:, 1] = y_caleb

    for j in range(48):
        all_coord[:, j + 2] = UMAP_noise * np.random.randn(len(x_caleb))
    adata = sc.AnnData(ssp.csr_matrix(all_coord))
    adata.obsm["X_clone"] = clone_annot
    adata.obs["time_info"] = time_info
    adata.obs["state_info"] = state_annote_0.astype(int).astype(str)
    adata.obsm["X_emb"] = all_coord[:, :2]
    adata.uns["data_des"] = ["bifurcation"]
    adata.obsm["X_orig"] = final_coordinates
    return adata


def linear_differentiation_model(
    Nt1=400,
    progeny_N=1,
    L=100,
    diff_sigma=0.5,
    dL=1,
    display=1,
    always_simulate_data=0,
    used_clone_N=100,
):
    """
    Simulate linear differentiation corrupted with barcode collision (See Fig. 3a)

    Parameters
    ----------
    Nt1:
        Number of initial cell states. They are randomly sampled along the differentiation manifold. Default 400.
    progeny_N:
        Number of progeny for each initial cell state
    L:
        The total progression length.
    diff_sigma:
        Differentiation noise.
    dL:
        Unit progression in a single step
    display:
        Plot figures or not
    always_simulate_data:
        Simulate new data (do not load pre-saved datasets)

    Returns
    -------
    adata:
        An adata object with clonal matrix, time info etc. Ready to be plugged into CoSpar.
    """

    if always_simulate_data:
        file_name = "simulate_data"
    else:
        file_name = f"{settings.data_path}/simulated_clonal_data_Nt1{Nt1}_progeny{progeny_N}_L{L}_dL{dL}_diffSigma{diff_sigma}"
    if os.path.exists(file_name + "_clonal_annot.npz"):
        print("Load existing data")
        clone_annot = ssp.load_npz(file_name + "_clonal_annot.npz")
        simu_data = np.load(file_name + "_others.npz", allow_pickle=True)
        x_t1 = simu_data["x_t1"]
        x_t2 = simu_data["x_t2"]
        Nt2 = simu_data["Nt2"]
        new_id_matrix = simu_data["new_id_matrix"]
        n = Nt1 * 20  # resolution of grid for sampling
        dx = L / n
    else:
        print("Generate new data")
        t = time.time()
        ## simulate the clonal data for each initial cell
        n = Nt1 * 20  # resolution of grid for sampling
        dx = L / n
        lattice = np.linspace(0, L, n)
        t1_id = np.sort(
            np.random.choice(n, Nt1, replace=False)
        )  # random sample of indices
        x_t1 = lattice[t1_id]
        id_matrix = np.zeros(
            (Nt1, progeny_N), dtype=int
        )  # information for single cell clones
        for j, x in enumerate(x_t1):
            ## this method is much faster
            if x < L - (1.5 * progression(x, dL) + 5 * sigma(x, diff_sigma)):
                bb = np.random.normal(
                    x + progression(x, dL), sigma(x, diff_sigma), progeny_N
                )
                id_matrix[j] = np.floor(bb / dx).astype(int)
            else:
                ## this method is much slower
                for k in range(progeny_N):
                    cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, dL))
                    # pick the first event that is larger than the random variable
                    id_matrix[j, k] = np.nonzero(cum_Prob > np.random.rand())[0][
                        0
                    ].astype(int)

        ## prepare clonal data
        t2_id = np.sort(id_matrix.flatten())
        x_t2 = np.sort(lattice[t2_id])
        Nt2 = len(t2_id)

        ## generate clonal matrix
        new_id_matrix = np.zeros(id_matrix.shape, dtype=int)
        # for j in range(id_matrix.shape[0]):
        for j in range(id_matrix.shape[0]):
            for k in range(id_matrix.shape[1]):
                # print(id_matrix[j,k])
                new_id_matrix[j, k] = Nt1 + np.nonzero(t2_id == id_matrix[j, k])[0][0]

        clone_annot = np.zeros((Nt1 + Nt2, Nt1), dtype=int)
        for j in range(Nt1):
            clone_annot[j, j] = 1
            clone_annot[new_id_matrix[j, :], j] = 1

        clone_annot = ssp.csr_matrix(clone_annot)

        # if display:
        #     idx = 3
        #     # cum_Prob=np.cumsum(transition_prob(x_t1[idx],lattice),dL)
        #     norm_prob = transition_prob(x_t1[idx], lattice, diff_sigma, dL)
        #     norm_prob = norm_prob / np.max(norm_prob)

        #     fig = plt.figure(figsize=(4, 3.5))
        #     ax = fig.add_subplot(1, 1, 1)
        #     ax.plot(lattice, norm_prob, label="Relative Prob.")
        #     x0 = lattice[id_matrix[idx]]
        #     y0 = np.array([1, 1, 1])
        #     ax.plot(x0, y0, ".r", label="Sampled point")
        #     leg = ax.legend()
        #     ax.set_xlabel("Differentiation progression")
        #     plt.tight_layout()
        #     fig.savefig(f"{settings.figure_path}/sample_data.eps")

        if not always_simulate_data:
            ssp.save_npz(file_name + "_clonal_annot.npz", clone_annot)
            np.savez(
                file_name + "_others.npz",
                x_t1=x_t1,
                x_t2=x_t2,
                Nt1=Nt1,
                Nt2=Nt2,
                new_id_matrix=new_id_matrix,
            )
        print("Time elapsed for generating clonal data: ", time.time() - t)

    ## cell annotation
    x_tot = np.array(list(x_t1) + list(x_t2))
    time_index_1 = np.array(range(Nt1))
    time_index_2 = np.array(range(Nt1, Nt1 + Nt2))
    time_info = np.zeros(len(x_tot))
    time_info[time_index_1] = 1
    time_info[time_index_2] = 2

    ## Compress the clones via random mixing
    compressed_N = np.round(Nt1 / used_clone_N).astype(int)
    CS_file_name = "generate new ones"
    if os.path.exists(CS_file_name):
        print("Load existing mixing matrix")
        mixing_matrix = np.load(CS_file_name)
    else:
        if display:
            print("Generate mixing matrix")
        N_cell, N_clone = clone_annot.shape

        post_PC_clone_N = np.floor(N_clone / compressed_N).astype(int)
        mixing_matrix = np.zeros((N_clone, post_PC_clone_N))
        current_clone_list = list(range(N_clone))
        for j in range(post_PC_clone_N):
            ri = np.random.choice(
                len(current_clone_list), compressed_N, replace=False
            )  # random sample of indices
            clone_ids = np.array(current_clone_list)[ri]
            mixing_matrix[clone_ids, j] = 1
            # remove these selected elements in the list
            for k in range(compressed_N):
                current_clone_list.remove(clone_ids[k])

    clone_annot_CS = ssp.csr_matrix(clone_annot * mixing_matrix)

    ## transformation to 50-d
    UMAP_noise = 0.2  #
    all_coord = np.zeros((len(x_tot), 50))
    all_coord[:, 0] = x_tot
    for j in range(49):
        all_coord[:, j + 1] = UMAP_noise * np.random.randn(len(x_tot))

    ## generate the adata
    adata = sc.AnnData(ssp.csr_matrix(all_coord))
    adata.obsm["X_clone"] = clone_annot_CS
    adata.obs["time_info"] = time_info
    adata.obs["state_info"] = np.zeros(len(x_tot)).astype(str)
    X_emb = np.zeros((len(x_tot), 2))
    X_emb[:, 0] = x_tot
    X_emb[:, 1] = x_tot
    adata.obsm["X_emb"] = X_emb
    adata.uns["data_des"] = ["linear_bifurcation"]
    adata.obsm["X_orig"] = x_tot
    return adata


def quantify_correlation_with_ground_truth_fate_bias_BifurcationModel(
    used_map, state_annote_0, cell_id_t1, cell_id_t2
):
    """
    Quantify the correlation of an inferred fate bias with the actual one, using a map from the Bifurcation Model.

    Parameters
    ----------
    used_map:
        Used tmap
    state_annote_0:
        The actual state annotation outputed from the Bifurcation Model.
    cell_id_t1:
        List of initial cell IDs
    cell_id_t2:
        List of later cell IDs

    Returns
    -------
    fate_bias_corr:
        Correlation between the predicted and actual fate bias
    """

    state_annote = state_annote_0[cell_id_t2]
    fate_array = ["0", "1", "-1"]

    potential_vector, fate_entropy = tl.compute_state_potential(
        used_map, state_annote, fate_array, fate_count=False
    )

    potential_vector = potential_vector + 0.01
    fate_bias = potential_vector[:, 0] / (
        potential_vector[:, 0] + potential_vector[:, 1]
    )

    state_annote_1 = state_annote_0[cell_id_t1]
    expected_fate_prob = np.zeros(len(state_annote_1))
    actual_fate_prob = np.zeros(len(state_annote_1))

    idx = state_annote_1 == "-1"
    expected_fate_prob[idx] = 0.5  # neutral
    actual_fate_prob[idx] = fate_bias[idx]

    idx = state_annote_1 == "0"
    expected_fate_prob[idx] = 1  # one branch
    actual_fate_prob[idx] = fate_bias[idx]

    idx = state_annote_1 == "1"
    expected_fate_prob[idx] = 0  # the other branch
    actual_fate_prob[idx] = fate_bias[idx]

    fate_bias_corr = np.corrcoef(expected_fate_prob, actual_fate_prob)[0, 1]

    return fate_bias_corr


def quantify_transition_peak_TPR_LinearDifferentiation(
    used_map, x_t1, x_t2, dL=1, relative_tolerance=3, diff_sigma=0.5, display=True
):
    """
    Quantify the True positive rate of a transition map for linear differentiation model.

    Parameters
    ----------
    used_map:
        Used tmap, in the form of numpy array, or sparse matrix
    x_t1:
        List of initial cell positions
    x_t2:
        List of later cell positions
    dL:
        Unit progression in a single time step
    relative_tolerance:
        Accept a prediction to be true if its most likely predicted target state is within
        relative_tolerance*diff_sigma of the expected target state.
    diff_sigma:
        Strength of differentiation noise.
    display:
        Show analysis figures if true.

    Returns
    --------
    Tmap_right_ratio:
        True positive rate of a transition map
    """

    if ssp.issparse(used_map):
        used_map = used_map.A

    Nt1 = len(x_t1)
    true_x = np.zeros(Nt1)
    predicted_x = np.zeros(Nt1)
    Tmap_right_ratio = []

    for ini_id in range(Nt1):
        true_x[ini_id] = x_t1[ini_id] + progression(x_t1[ini_id], dL)

        temp_map = used_map[ini_id]
        idx = temp_map > 0.999 * np.max(temp_map)
        if np.sum(idx) > 1:
            temp_sel_idx = np.random.choice(np.nonzero(idx)[0], 1)[0]
            predicted_temp_x = x_t2[temp_sel_idx]
        else:
            predicted_temp_x = x_t2[np.argsort(temp_map)[::-1][0]]

        predicted_x[ini_id] = predicted_temp_x

    relative_x = predicted_x - true_x
    Tmap_right_ratio = (
        1 - np.sum(abs(relative_x) > relative_tolerance * diff_sigma) / Nt1
    )

    if display:
        fig = plt.figure(figsize=(4, 3.5))
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(predicted_x, true_x, ".r", label="ICSLAM")
        ax.set_xlabel("Predicted progression: $x_2$")
        # ax.legend(loc='best')
        ax.set_ylabel("Actual progression: $x_2$")
        # ax.set_title(f'Ini. cell #={Nt1}, Clone #={clone_annot_CS.shape[1]}')
        plt.tight_layout()
        fig.savefig(f"{settings.figure_path}/progression_comparison_1.eps")

    return Tmap_right_ratio



# -*- coding: UTF-8 -*-
import matplotlib
matplotlib.use('Agg')
import os
import random
import anndata
import numpy as np
import scanpy as sc
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from scipy.sparse import find
from scipy.sparse import csr_matrix



def plot_Clonemap(X_clone, C_data, node_size, save_path, 
                  with_labels=False, draw_edges=True, 
                  specified_tree_index=None, specified_node_name=None,
                  plot_clusters=False):
    
    xclone_path = os.path.join(save_path, 'barcode_singlecell')
    xclone_cluster_path = os.path.join(save_path, 'barcode_cluster')
    

    #X_clone是读取h5ad的anndata里的adata.obsm['X_clone']
    #C_data是读取聚类数据csv，也就是pd.read_csv('combined_monocle2.csv')而combined_monocle2.csv是matlab里对count_result_rank直接使用writetable存的文件
    #node_size是所作图像的点的大小
    #save_path是所作图像的存储路径
    sorted_C_data = C_data.sort_values(by='Var3')
    cell_names = sorted_C_data['Var3'].values
    X1 = sorted_C_data['Var1'].values
    X2 = sorted_C_data['Var2'].values
    T = sorted_C_data['Pst'].values

    # 染色方案
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_gray_blue", ["#17324b", "#61ade8"])
    
    index_map = {name: idx for idx, name in enumerate(cell_names)}
    rev_index_map = {idx: name for idx, name in enumerate(cell_names)}
    # 创建伪时间数组
    lenT = len(T)
    # 创建有向图
    G = nx.DiGraph()

    # 从稀疏矩阵中找出非零元素的位置
    rows, cols, _ = find(X_clone)  

    # 将稀疏矩阵逐点提取，以细胞索引为点，逐个添加有向边
    for row, col in zip(rows, cols):
        if row < lenT and col < lenT:
            if T[row] < T[col]:
                G.add_edge(row, col)
            else:
                G.add_edge(col, row)

    # 识别所有弱连通分量
    weak_components = list(nx.weakly_connected_components(G))
    forest = []

    # 遍历每个弱连通分量
    for component in weak_components:
        subgraph = G.subgraph(component).copy()
        roots = [node for node in subgraph if subgraph.in_degree(node) == 0]
        if not roots:
            roots = [next(iter(component))]

        for root in roots:
            tree = nx.dfs_tree(subgraph, source=root)
            forest.append(tree)

    # 确保保存路径存在
    os.makedirs(save_path, exist_ok=True)

    # 根据指定的参数绘制图像

    trees_to_plot = []
    original_indices = []

    if specified_tree_index is not None:
        if isinstance(specified_tree_index, int):
            specified_tree_index = [specified_tree_index]
        trees_to_plot = [forest[i] for i in specified_tree_index if i < len(forest)]
        original_indices = [i for i in specified_tree_index if i < len(forest)]
    elif specified_node_name is not None:
        if isinstance(specified_node_name, int):
            specified_node_name = [specified_node_name]
        node_indices = [index_map.get(name) for name in specified_node_name if name in index_map]
        for index, tree in enumerate(forest):
            if any(node in tree.nodes for node in node_indices):
                trees_to_plot.append(tree)
                original_indices.append(index)
    else:
        trees_to_plot = forest
        original_indices = list(range(len(forest)))

    all_trees = len(trees_to_plot)
    # 绘制指定的树
    if plot_clusters == True:
        for tree in trees_to_plot:
            nodes_in_tree = [node for node in tree.nodes()]
            node_names_in_tree = [rev_index_map[node] for node in nodes_in_tree if node in rev_index_map]
            #print(node_names_in_tree)
            plot_Pseudotimemap(C_data, node_size, save_path, specified_cell_name=node_names_in_tree,plot_clusters=True,use_pseudotime=True,go_on=1)
    else:
        for i, (index, tree) in enumerate(zip(original_indices, trees_to_plot), start=1):
            print(f"Plotting... ( {i} / {all_trees} )", end='\r')  # 输出当前进度
            plt.figure(figsize=(16, 9))
            plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)
            pos = {node: (X1[node], X2[node]) for node in tree.nodes() if node < lenT}
            node_colors = [T[node] for node in tree.nodes() if node < lenT]
            labels = {node: rev_index_map[node] for node in tree.nodes() if node in rev_index_map}
            if draw_edges:
                nc = nx.draw(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap,
                       labels=labels if with_labels else None, edge_color='gray', node_size=node_size, font_size=8)
            else:
                nc = nx.draw_networkx_nodes(tree.subgraph(pos.keys()), pos, node_color=node_colors, cmap=cmap, node_size=node_size)

            # 使用nc作为mappable对象传递给colorbar
            plt.colorbar(nc, orientation='vertical', label='Pseudotime')
            plt.title(f"Barcode tracing - Tree {index}")
            plt.xlabel('UMAP 1')
            plt.ylabel('UMAP 2')
            plt.savefig(os.path.join(xclone_path, f'Tree_{index}.png'))
            plt.close()




def plot_Pseudotimemap(C_data, node_size, save_path, Color="#618a03",
                 specified_cluster_index=None,specified_cell_name=None,
                 with_labels=False,plot_clusters=True,use_pseudotime=False,go_on=0):
    #C_data是读取聚类数据csv，也就是pd.read_csv('combined_monocle2.csv')
    #而combined_monocle2.csv是matlab里对count_result_rank直接使用writetable存的文件
    #node_size是所作图像的点的大小
    #save_path是所作图像的存储路径
    print('Pseudotimemap')
    pseudotime_path = os.path.join(save_path, 'pseudotime_singlecell')
    pseudotime_cluster_path = os.path.join(save_path, 'pseudotime_cluster')
    xclone_cluster_path = os.path.join(save_path, 'barcode_cluster')
    

    X1 = C_data['Var1'].values
    X2 = C_data['Var2'].values
    T  =  C_data['Pst'].values
    # 染色方案
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_gray_blue", ["#17324b", "#61ade8"])

    cellname = {name: idx for idx, name in enumerate(C_data['Var3'])}
    # 根据MM_new_rank分组细胞名到clusters
    clusters = C_data.groupby('MM_new_rank')['Var3'].apply(set).to_dict()

    plt.figure(figsize=(16, 9))
    # 绘制背景点
    plt.scatter(X1, X2, c='lightgray', s=node_size, alpha=0.6)

    # 处理输入参数，允许单个或多个输入
    if isinstance(specified_cluster_index, int):
        specified_cluster_index = [specified_cluster_index]
    if isinstance(specified_cell_name, int):
        specified_cell_name = [specified_cell_name]
    if isinstance(Color, str):
        Color = [Color]

    # 确定需要绘制的簇
    if specified_cell_name:
        specified_cluster_index = [idx for idx, cells in clusters.items()
                                   if any(node in cells for node in specified_cell_name)]
        filename_suffix = f'_cell{format_suffix(specified_cell_name)}'
    elif specified_cluster_index:
        filename_suffix = f'_cluster{format_suffix(specified_cluster_index)}'
    else:
        specified_cluster_index = list(clusters.keys())
        filename_suffix = '_all'
    #若预先指定的Color不够则扩展 Color 列表
    while len(Color) < len(specified_cluster_index):
        Color.append(f"#{random.randint(0, 0xFFFFFF):06x}")
    # 绘制指定的簇
    if plot_clusters:
        for i, idx in enumerate(specified_cluster_index):
            if idx in clusters:
                node_indices = [cellname[name] for name in clusters[idx]]
                if use_pseudotime:
                    color = T[node_indices]  # 使用伪时间的颜色
                    plt.scatter(X1[node_indices], X2[node_indices], c=color, cmap=cmap, s=node_size, label=f'Cluster {idx}')
                else:                        # 使用Color的颜色
                    plt.scatter(X1[node_indices], X2[node_indices], c=Color[i], s=node_size, label=f'Cluster {idx}')
    # 在图中特别标出 specified_cell_name 对应的节点
    if specified_cell_name:
        special_indices = [cellname[name] for name in specified_cell_name if name in cellname]
        if use_pseudotime:
            color = T[special_indices]  # 使用伪时间颜色
            plt.scatter(X1[special_indices], X2[special_indices], c=color, cmap=cmap, s=node_size+2, linewidths=1)
        else:
            plt.scatter(X1[special_indices], X2[special_indices], c='red', s=node_size+2, linewidths=1)
        if with_labels:
            for idx in special_indices:
                plt.text(X1[idx], X2[idx], f'{C_data["Var3"].iloc[idx]}', fontsize=9, color='black', ha='right')
    # 添加图例
    if plot_clusters:
        plt.legend(title='Clusters')
    plt.title("UMAP Plot" + (" with Cluster Coloring" if plot_clusters else "") + (" using Pseudotime" if use_pseudotime else ""))
    plt.xlabel('UMAP 1')
    plt.ylabel('UMAP 2')
    if plot_clusters == True and go_on == 0:
        plt.savefig(f'{pseudotime_cluster_path}\clusters_plot{filename_suffix}.png')
    if plot_clusters == False and go_on == 0:
        plt.savefig(f'{pseudotime_path}\clusters_plot{filename_suffix}.png')
    if plot_clusters == True and go_on == 1:
        print('Continue from Barcode cluster')
        plt.savefig(f'{xclone_cluster_path}\clusters_plot{filename_suffix}.png')
    plt.close()

def format_suffix(items):
    if len(items) > 4:
        formatted = f'{",".join(map(str, items[:3]))},...,{items[-1]}'
    else:
        formatted = ",".join(map(str, items))
    return formatted



'''
# 路径设置
file_path = 'Hematopoiesis_progenitor.h5ad'
cluster_path = 'combined_monocle2.csv'
# 读取H5AD
adata = anndata.read_h5ad(file_path)
# 获取X_clone矩阵
X_clone = adata.obsm['X_clone']
# 读取CSV
cluster_data = pd.read_csv(cluster_path)
node_size=8
save_path='G:/ASUMDownloads/dataplot2'
plot_Clusters(cluster_data, node_size, save_path,plot_clusters=True,specified_cell_name=9429,use_pseudotime=True)
'''
